
import gzip, os, csv, struct, zipfile, io
import logging

_sqlite_keywords = {
	'ABORT', 'ACTION', 'ADD', 'AFTER', 'ALL', 'ALTER',
	'ANALYZE', 'AND', 'AS', 'ASC', 'ATTACH', 'AUTOINCREMENT', 'BEFORE', 'BEGIN',
	'BETWEEN', 'BY', 'CASCADE', 'CASE', 'CAST', 'CHECK', 'COLLATE', 'COLUMN',
	'COMMIT', 'CONFLICT', 'CONSTRAINT', 'CREATE', 'CROSS', 'CURRENT_DATE',
	'CURRENT_TIME', 'CURRENT_TIMESTAMP', 'DATABASE', 'DEFAULT', 'DEFERRABLE',
	'DEFERRED', 'DELETE', 'DESC', 'DETACH', 'DISTINCT', 'DROP', 'EACH', 'ELSE',
	'END', 'ESCAPE', 'EXCEPT', 'EXCLUSIVE', 'EXISTS', 'EXPLAIN', 'FAIL', 'FOR',
	'FOREIGN', 'FROM', 'FULL', 'GLOB', 'GROUP', 'HAVING', 'IF', 'IGNORE',
	'IMMEDIATE', 'IN', 'INDEX', 'INDEXED', 'INITIALLY', 'INNER', 'INSERT',
	'INSTEAD', 'INTERSECT', 'INTO', 'IS', 'ISNULL', 'JOIN', 'KEY', 'LEFT',
	'LIKE', 'LIMIT', 'MATCH', 'NATURAL', 'NO', 'NOT', 'NOTNULL', 'NULL', 'OF',
	'OFFSET', 'ON', 'OR', 'ORDER', 'OUTER', 'PLAN', 'PRAGMA', 'PRIMARY', 'QUERY',
	'RAISE', 'REFERENCES', 'REGEXP', 'REINDEX', 'RELEASE', 'RENAME', 'REPLACE',
	'RESTRICT', 'RIGHT', 'ROLLBACK', 'ROW', 'SAVEPOINT', 'SELECT', 'SET',
	'TABLE', 'TEMP', 'TEMPORARY', 'THEN', 'TO', 'TRANSACTION', 'TRIGGER',
	'UNION', 'UNIQUE', 'UPDATE', 'USING', 'VACUUM', 'VALUES', 'VIEW', 'VIRTUAL',
	'WHEN', 'WHERE'
}



class SmartFileReader(object):
	def __init__(self, file, *args, **kwargs):
		if file[-3: ] =='.gz':
			with open(file, 'rb') as f:
				f.seek(-4, 2)
				self._filesize = struct.unpack('I', f.read(4))[0]
			self.file = gzip.open(file, 'rt', *args, **kwargs)
		elif file[-4: ] =='.zip':
			zf = zipfile.ZipFile(file, 'r')
			zf_info = zf.infolist()
			if len(zf_info ) !=1:
				raise TypeError("zip archive files must contain a single member file for SmartFileReader")
			zf_info = zf_info[0]
			self.file = zf.open(zf_info.filename, 'r', *args, **kwargs)
			self._filesize = zf_info.file_size
		else:
			self.file = open(file, 'rt', *args, **kwargs)
			self._filesize = os.fstat(self.file.fileno()).st_size
	def __getattr__(self, name):
		return getattr(self.file, name)
	def __setattr__(self, name, value):
		if name in ['file', 'percentread', '_filesize']:
			return object.__setattr__(self, name, value)
		return setattr(self.file, name, value)
	def __delattr__(self, name):
		return delattr(self.file, name)
	def percentread(self):
		try:
			return (float(self.file.tell() ) /float(self._filesize ) *100)
		except io.UnsupportedOperation:
			try:
				return 1.0- (float(self.file._left) / float(self._filesize) * 100)
			except:
				return 0

	def __iter__(self):
		return self.file.__iter__()

	def bytesread(self):
		try:
			b = float(self.file.tell())
		except:
			return "error in bytesread"
		labels = ['B', 'KB', 'MB', 'GB', 'TB']
		scale = 0
		while scale < 4 and b > 1024:
			b /= 1024
			scale += 1
		return "{:.2f}{}".format(b, labels[scale])

	def progress(self):
		pct = self.percentread()
		if pct > 0 and pct < 100:
			return "{}%".format(pct)
		else:
			return self.bytesread()


def interpret_header_name(h):
	eL = logging.getLogger("l4util")
	if h.lower() == "case" or h[1:].lower() == "case":
		eL.warning("'case' is a reserved keyword, changing column header to 'caseid'")
		h = h + "id"
	if h.lower() == "group" or h[1:].lower() == "group":
		eL.warning("'group' is a reserved keyword, changing column header to 'groupid'")
		h = h + "id"
	if h.upper() in _sqlite_keywords:
		eL.warning("'%s' is a reserved keyword, changing column header to '%s_'" % (h, h))
		h = h + "_"
	if '.' in h:
		h = h.replace('.', '_')
		eL.warning("dots are not allowed in column headers, changing column header to '%s'" % (h,))
	if len(h) == 0:
		eL.error("zero length header")
		return " "
	if h[0] == '@':
		return h[1:] + " TEXT"
	elif h[0] == '#':
		return h[1:] + " INTEGER"
	return h + " NUMERIC"


def prepare_import_headers(rawfilename, headers=None):
	'''
	Identify the file type and read the column headers from the first row.
	For each column, data type is presumed to be FLOAT unless the column name is
	prepended with '@' or '#' which makes the column format TEXT or INT respectively.
	Since data is stored in SQLite, format is just a suggestion anyhow.

	Input:
	  rawfile: the filename (absolute or relative path) to the text file with data.
	Output:
	  headers: a series of 2-tuples identifying column names and data types
	  r: the csv reader object, primed to begin reading at the second line of the file
	'''
	eL = logging.getLogger("l4util")
	if isinstance(rawfilename, str):
		raw = SmartFileReader(rawfilename)
	else:
		raw = rawfilename
	sample = raw.read(28192)
	raw.seek(0)
	if isinstance(sample, bytes):
		sample = sample.decode('UTF-8')
	try:
		dialect = csv.Sniffer().sniff(sample)
	except csv.Error:
		sample = raw.read(512000)
		raw.seek(0)
		if isinstance(sample, bytes):
			sample = sample.decode('UTF-8')
		dialect = csv.Sniffer().sniff(sample)
	r = csv.reader(raw, dialect)
	eL.debug("DIALECT = %s", str(dialect))
	eL.debug("TYPE OF READER = %s", str(type(r)))
	if headers is None:
		try:
			headers = r.next()
		except AttributeError:
			headers = next(r)  ##
	# fix header items for type
	for h in range(len(headers)):
		headers[h] = interpret_header_name(headers[h])
	return headers, r, raw


def prepare_import_headers_plain(rawfilename, headers=None):
	'''
	Identify the file type and read the column headers from the first row.
	For each column, data type is presumed to be FLOAT unless the column name is
	prepended with '@' or '#' which makes the column format TEXT or INT respectively.
	Since data is stored in SQLite, format is just a suggestion anyhow.

	Input:
	  rawfile: the filename (absolute or relative path) to the text file with data.
	Output:
	  headers: a series of 2-tuples identifying column names and data types
	  r: the csv reader object, primed to begin reading at the second line of the file
	'''
	eL = logging.getLogger("l4util")
	if isinstance(rawfilename, str):
		raw = SmartFileReader(rawfilename)
	else:
		raw = rawfilename
	sample = raw.read(28192)
	raw.seek(0)
	if isinstance(sample, bytes):
		sample = sample.decode('UTF-8')
	try:
		dialect = csv.Sniffer().sniff(sample)
	except csv.Error:
		sample = raw.read(512000)
		raw.seek(0)
		if isinstance(sample, bytes):
			sample = sample.decode('UTF-8')
		dialect = csv.Sniffer().sniff(sample)
	r = csv.reader(raw, dialect)
	eL.debug("DIALECT = %s", str(dialect))
	eL.debug("TYPE OF READER = %s", str(type(r)))
	if headers is None:
		try:
			headers = r.next()
		except AttributeError:
			headers = next(r)  ##
	return headers, r, raw



import unicodedata
from difflib import SequenceMatcher as _SequenceMatcher
from heapq import nlargest as _nlargest
from pathlib import Path


def strip_accents(s):
	return ''.join(c for c in unicodedata.normalize('NFD', s)
	               if unicodedata.category(c) != 'Mn')


def supercasefold(s):
	return strip_accents(s).casefold()


def case_insensitive_close_matches(word, possibilities, n=3, cutoff=0.6, excpt=None):
	"""Use SequenceMatcher to return list of the best "good enough" matches.

	word is a sequence for which close matches are desired (typically a
	string).

	possibilities is a list of sequences against which to match word
	(typically a list of strings).

	Optional arg n (default 3) is the maximum number of close matches to
	return.  n must be > 0.

	Optional arg cutoff (default 0.6) is a float in [0, 1].  Possibilities
	that don't score at least that similar to word are ignored.

	The best (no more than n) matches among the possibilities are returned
	in a list, sorted by similarity score, most similar first.

	>>> get_close_matches("appel", ["ape", "apple", "peach", "puppy"])
	['apple', 'ape']
	>>> import keyword as _keyword
	>>> get_close_matches("wheel", _keyword.kwlist)
	['while']
	>>> get_close_matches("Apple", _keyword.kwlist)
	[]
	>>> get_close_matches("accept", _keyword.kwlist)
	['except']
	"""

	if not n > 0:
		raise ValueError("n must be > 0: %r" % (n,))
	if not 0.0 <= cutoff <= 1.0:
		raise ValueError("cutoff must be in [0.0, 1.0]: %r" % (cutoff,))
	result = []
	s = _SequenceMatcher()
	s.set_seq2(supercasefold(word))
	for x in possibilities:
		x_ = supercasefold(x)
		s.set_seq1(x_)
		if s.real_quick_ratio() >= cutoff and \
						s.quick_ratio() >= cutoff and \
						s.ratio() >= cutoff:
			result.append((s.ratio(), x))

	# Move the best scorers to head of list
	result = _nlargest(n, result)
	# Strip scores for the best n matches
	ret = [x for score, x in result]
	if not excpt is None:
		did_you_mean = "'{}' not found, did you mean {}?".format(word, " or ".join("'{}'".format(s) for s in ret))
		raise excpt(did_you_mean)
	return ret


def max_len(arg, at_least=0):
	if len(arg) == 0:
		return at_least
	try:
		return max(at_least, *map(len, arg))
	except:
		try:
			return max(at_least, *map(len, *arg))
		except:
			print("ERR max_len")
			print("arg=", arg)
			print("type(arg)=", type(arg))
			raise


def grid_str(arg, lineprefix="", linesuffix=""):
	x = []
	try:
		n = 0
		while 1:
			x.append(max_len(tuple(i[n] for i in arg)))
			n += 1
	except IndexError:
		n -= 1
	lines = [lineprefix + " ".join(t[i].ljust(x[i]) for i in range(len(t))) + linesuffix for t in arg]
	return '\n'.join(lines)



def truncate_path_for_display(p, maxlen=30):
	"""Truncates a path if it is too long to display neatly.

	Do not use the trunacted path for actual file operations, it won't work.

	Parameters
	----------
	p : str or pathlib.Path
	maxlen : int
		Maximum number of character to return.  This is a suggestion, as
		the basename itself is never truncated.
	"""
	original = Path(p)
	if len(str(original)) <= maxlen:
		return str(original)
	original_parts_len = len(original.parts)
	trimmer = 1
	z = lambda x: str(Path('⋯', *original.parts[x:]))    # unicode equivalent '\u22EF'
	while len(z(trimmer)) > maxlen:
		trimmer += 1
		if trimmer >= original_parts_len:
			return z(original_parts_len-1)
	return z(trimmer)




def stable_df(df, filename):
	import os, pandas
	from pandas.io.formats.style import Styler
	if isinstance(df, Styler):
		df = df.data
	f = os.path.join(os.path.dirname(__file__), f'{filename}.pkl.gz')
	if os.path.exists(f):
		comparison = pandas.read_pickle(f)
		pandas.testing.assert_frame_equal(df, comparison)
	else:
		try:
			df.to_pickle(f)
		except FileNotFoundError:
			raise FileNotFoundError(f)





import numpy
import pandas

from .util import Dict, dictx
from .model import Model

import logging
from .log import logger_name
logger = logging.getLogger(logger_name)

def doctor(
		dfs,
		repair_ch_av=None,
		repair_ch_zq=None,
		repair_asc=None,
		repair_noch_nowt=None,
		repair_nan_wt=None,
		verbose=3,
):
	problems = dictx()

	if isinstance(dfs, Model) and dfs.dataframes is None:
		raise ValueError('no dataframes loaded, try `.load_data()` first')

	logger.info("checking for chosen-but-zero-quantity")
	dfs, diagnosis = chosen_but_zero_quantity(dfs, repair=repair_ch_zq, verbose=verbose)
	if diagnosis is not None:
		logger.warning(f'problem: chosen-but-zero-quantity ({len(diagnosis)} issues)')
		problems['chosen_but_zero_quantity'] = diagnosis

	logger.info("checking for chosen-but-not-available")
	dfs, diagnosis = chosen_but_not_available(dfs, repair=repair_ch_av, verbose=verbose)
	if diagnosis is not None:
		logger.warning(f'problem: chosen-but-not-available ({len(diagnosis)} issues)')
		problems['chosen_but_not_available'] = diagnosis

	logger.info("checking for nothing-chosen-but-nonzero-weight")
	dfs, diagnosis = nothing_chosen_but_nonzero_weight(dfs, repair=repair_noch_nowt, verbose=verbose)
	if diagnosis is not None:
		logger.warning(f'problem: nothing-chosen-but-nonzero-weight ({len(diagnosis)} issues)')
		problems['nothing_chosen_but_nonzero_weight'] = diagnosis

	logger.info("checking for nan-weight")
	dfs, diagnosis = nan_weight(dfs, repair=repair_nan_wt, verbose=verbose)
	if diagnosis is not None:
		logger.warning(f'problem: nan-weight ({len(diagnosis)} issues)')
		problems['nan_weight'] = diagnosis

	logger.info("checking for low-variance-data-co")
	dfs, diagnosis = low_variance_data_co(dfs, repair=None, verbose=verbose)
	if diagnosis is not None:
		logger.warning(f'problem: low-variance-data-co ({len(diagnosis)} issues)')
		problems['low_variance_data_co'] = diagnosis

	# if repair_asc:
	# 	x = self.cleanup_asc_problems()
	# 	if x is not None:
	# 		problems['asc_for_never_chosen'] = x
	return dfs, problems



def chosen_but_not_available(dfs, repair=None, verbose=3):
	"""
	Check if some observations are chosen but not available

	Parameters
	----------
	dfs : DataFrames or Model
		The data to check
	repair : {None, '+', '-', }
		How to repair the data.
		Plus will make the conflicting alternatives available.
		Minus will make them not chosen (possibly leaving no chosen alternative).
		None effects no repair, and simply emits a warning.
	verbose : int, default 3
		The number of example rows to list for each problem.

	Returns
	-------
	dfs : DataFrames
		The revised dataframe
	diagnosis : pandas.DataFrame
		The number of bad instances, by alternative, and some example rows.

	"""

	if isinstance(dfs, Model):
		m = dfs
		dfs = m.dataframes
	else:
		m = None

	if m is not None and dfs.data_ch.shape[1] == len(m.graph):
		# choice data is wide, make availability data wide
		data_av = dfs.data_av_cascade(m.graph)
	else:
		data_av = dfs.data_av

	_not_avail = (data_av==0).values
	_chosen = (dfs.data_ch > 0).values
	_wid = min(_not_avail.shape[1], _chosen.shape[1])

	chosen_but_not_available = pandas.DataFrame(
		data=_not_avail[:, :_wid] & _chosen[:, :_wid],
		index=data_av.index,
		columns=data_av.columns[:_wid],
	)
	chosen_but_not_available_sum = chosen_but_not_available.sum(0)

	diagnosis = None
	if chosen_but_not_available_sum.sum() > 0:

		i1, i2 = numpy.where(chosen_but_not_available)

		diagnosis = pandas.DataFrame(
			chosen_but_not_available_sum[chosen_but_not_available_sum > 0],
			columns=['n', ],
		)

		for colnum, colname in enumerate(chosen_but_not_available.columns):
			if chosen_but_not_available_sum[colname] > 0:
				diagnosis.loc[colname, 'example rows'] = ", ".join(str(j) for j in i1[i2 == colnum][:verbose])

		if repair == '+':
			dfs.data_av.values[chosen_but_not_available] = 1
		elif repair == '-':
			dfs.data_ch.values[chosen_but_not_available] = 0

	if m is None:
		return dfs, diagnosis
	else:
		return m, diagnosis



def chosen_but_zero_quantity(dfs, repair=None, verbose=3):
	"""
	Check if some observations are chosen but have zero quantity.

	Parameters
	----------
	dfs : DataFrames or Model
		The data to check
	repair : {None, '-', }
		How to repair the data.
		Minus will make them not chosen (possibly leaving no chosen alternative).
		None effects no repair, and simply emits a warning.
	verbose : int, default 3
		The number of example rows to list for each problem.

	Returns
	-------
	dfs : DataFrames
		The revised dataframe
	diagnosis : pandas.DataFrame
		The number of bad instances, by alternative, and some example rows.

	"""

	if repair not in ('-', None):
		raise ValueError(f'invalid repair setting "{repair}"')

	if isinstance(dfs, Model):
		m = dfs
		dfs = m.dataframes
	else:
		m = None

	try:
		zero_q = dfs.get_zero_quantity_ca()
	except ValueError:
		diagnosis = None

	else:
		_zero_q = (zero_q>0).values
		_chosen = (dfs.data_ch > 0).values
		_wid = min(_zero_q.shape[1], _chosen.shape[1])

		chosen_but_zero_quantity = pandas.DataFrame(
			data=_zero_q[:, :_wid] & _chosen[:, :_wid],
			index=dfs.data_av.index,
			columns=dfs.data_av.columns[:_wid],
		)
		chosen_but_zero_quantity_sum = chosen_but_zero_quantity.sum(0)

		diagnosis = None
		if chosen_but_zero_quantity_sum.sum() > 0:

			i1, i2 = numpy.where(chosen_but_zero_quantity)

			diagnosis = pandas.DataFrame(
				chosen_but_zero_quantity_sum[chosen_but_zero_quantity_sum > 0],
				columns=['n', ],
			)

			for colnum, colname in enumerate(chosen_but_zero_quantity.columns):
				if chosen_but_zero_quantity_sum[colname] > 0:
					diagnosis.loc[colname, 'example rows'] = ", ".join(str(j) for j in i1[i2 == colnum][:verbose])

			if repair == '-':
				dfs.data_ch.values[chosen_but_zero_quantity] = 0

	if m is None:
		return dfs, diagnosis
	else:
		return m, diagnosis




def nothing_chosen_but_nonzero_weight(dfs, repair=None, verbose=3):
	"""
	Check if some observations are chosen but not available

	Parameters
	----------
	dfs : DataFrames or Model
		The data to check
	repair : {None, '-', '*'}
		How to repair the data.
		Minus will make the weight zero.
		Star will make the weight zero plus autoscale all remaining weights.
		None effects no repair, and simply emits a warning.
	verbose : int, default 3
		The number of example rows to list for each problem.

	Returns
	-------
	dfs : DataFrames
		The revised dataframe
	diagnosis : pandas.DataFrame
		The number of bad instances, by alternative, and some example rows.

	"""

	if isinstance(dfs, Model):
		m = dfs
		dfs = m.dataframes
	else:
		m = None

	diagnosis = None

	if dfs is None:
		raise ValueError('data not loaded')

	if dfs.data_wt is not None and dfs.data_ch is not None:

		nothing_chosen = (dfs.array_ch().sum(1) == 0)
		nothing_chosen_some_weight = nothing_chosen & (dfs.array_wt().reshape(-1) > 0)
		if nothing_chosen_some_weight.sum() > 0:

			i1 = numpy.where(nothing_chosen_some_weight)[0]

			diagnosis = pandas.DataFrame(
				[nothing_chosen_some_weight.sum(), ],
				columns=['n', ],
				index=['nothing_chosen_some_weight', ]
			)

			diagnosis.loc['nothing_chosen_some_weight', 'example rows'] = ", ".join(str(j) for j in i1[:verbose])
			if repair == '+':
				raise ValueError("cannot resolve chosen_but_zero_quantity by assuming some choice")
			elif repair == '-':
				dfs.array_wt()[nothing_chosen] = 0
			elif repair == '*':
				dfs.array_wt()[nothing_chosen] = 0
				dfs.autoscale_weights()
	if m is None:
		return dfs, diagnosis
	else:
		return m, diagnosis


def nan_weight(dfs, repair=None, verbose=3):
	"""
	Check if some observations are chosen but not available

	Parameters
	----------
	dfs : DataFrames or Model
		The data to check
	repair : None or bool
		Whether to repair the data.
		Any true value will make NaN values in the weight zero.
		None effects no repair, and simply emits a warning.
	verbose : int, default 3
		The number of example rows to list for each problem.

	Returns
	-------
	dfs : DataFrames
		The revised dataframe
	diagnosis : pandas.DataFrame
		The number of bad instances, and some example rows.

	"""
	if isinstance(dfs, Model):
		m = dfs
		dfs = m.dataframes
	else:
		m = None

	diagnosis = None
	if dfs.data_wt is not None:
		nan_wgt = numpy.isnan(dfs.data_wt.iloc[:, 0])

		if nan_wgt.sum():
			i = numpy.where(nan_wgt)[0]

			diagnosis = pandas.DataFrame(
				data=[[nan_wgt.sum(), '']],
				columns=['n', 'example rows'],
				index=['nan_weight'],
			)

			diagnosis.loc['nan_weight', 'example rows'] = ", ".join(str(j) for j in i[:verbose])

		if repair:
			dfs.data_wt.fillna(0, inplace=True)

	if m is None:
		return dfs, diagnosis
	else:
		return m, diagnosis



def low_variance_data_co(dfs, repair=None, verbose=3):
	"""
	Check if any data_co columns have very low variance.

	Parameters
	----------
	dfs : DataFrames or Model
		The data to check
	repair : None
		Not implemented.
	verbose : int, default 3
		The number of example columns to list if
		there is a problem.

	Returns
	-------
	dfs : DataFrames
		The revised dataframe
	diagnosis : pandas.DataFrame
		The number of bad instances, and some example rows.

	"""
	if isinstance(dfs, Model):
		m = dfs
		dfs = m.dataframes
	else:
		m = None

	diagnosis = None
	if dfs.data_co is not None:
		variance = dfs.data_co.var()
		if variance.min() < 1e-3:
			i = numpy.where(variance < 1e-3)[0]

			diagnosis = pandas.DataFrame(
				data=[[len(i), '']],
				columns=['n', 'example cols'],
				index=['low_variance_co'],
			)

			diagnosis.loc['low_variance_co', 'example cols'] = ", ".join(str(dfs.data_co.columns[j]) for j in i[:verbose])

	if m is None:
		return dfs, diagnosis
	else:
		return m, diagnosis



import larch
import numpy as np
import pandas as pd

from larch import P, X
from larch.model.constraints import RatioBound, OrderingBound, FixedBound
from larch.model.constraints import interpret_contraint
from pytest import approx
import pytest
import sys
from .stored_dataframes import stable_df

def test_max_ratio_1():

	#m0 = larch.Model.Example(1)
	m1 = larch.Model.Example(1)

	#m0.utility_ca = P.tottime * X.tottime + P.tottime * X.totcost
	#m0.remove_unused_parameters()

	#m0.set_value(P.tottime, maximum=0)
	m1.set_value(P.totcost, maximum=0)

	c1 = RatioBound(P.tottime, P.totcost, min_ratio=0.75, max_ratio=1.0, scale=1)

	#m0.load_data()
	m1.load_data()

	m1.constraints = [c1,]

	# r0 = m0.maximize_loglike(
	# 	method='slsqp',
	# )
	r1 = m1.maximize_loglike(
		method='slsqp',
	)

	assert r1['loglike'] == approx(-3764.1459333329217, rel=1e-4)
	assert 'Optimization terminated successfully' in r1.message
	assert m1['tottime'].value / m1['totcost'].value == approx(1.0, rel=1e-3)
	assert c1.fun(m1.pf.value) == approx(0, abs=1e-5)

	# m0.calculate_parameter_covariance()
	m1.calculate_parameter_covariance()

	# assert m0.pf.loc['tottime', 'std_err'] == approx(0.000247, rel=5e-3)
	assert m1.pf.loc['tottime', 'std_err'] == approx(0.000247, rel=5e-3)
	assert m1.pf.loc['totcost', 'std_err'] == approx(0.000247, rel=5e-3)

def test_max_ratio_2():

	# m0 = larch.Model.Example(1)
	m1 = larch.Model.Example(1)

	# m0.utility_ca = P.tottime * (X.tottime * 2) + P.tottime * X.totcost
	# m0.remove_unused_parameters()

	# m0.set_value(P.tottime, maximum=0)
	m1.set_value(P.totcost, maximum=0)

	c1 = RatioBound(P.tottime, P.totcost, min_ratio=0.75, max_ratio=2.0, scale=1)

	# m0.load_data()
	m1.load_data()

	m1.constraints = [c1,]

	# r0 = m0.maximize_loglike(
	# 	method='slsqp',
	# )
	r1 = m1.maximize_loglike(
		method='slsqp',
	)

	assert m1['tottime'].value / m1['totcost'].value == approx(2.0, rel=1e-3)

	assert r1['loglike'] == approx(-3730.8942681224476)

	assert 'Optimization terminated successfully' in r1.message
	assert c1.fun(m1.pf.value) == approx(0, abs=1e-5)

	# m0.calculate_parameter_covariance()
	m1.calculate_parameter_covariance()

	# assert m0.pf.loc['tottime', 'std_err'] == approx(0.000248, rel=5e-3)
	assert m1.pf.loc['tottime', 'std_err'] == approx(0.000248*2, rel=5e-3)
	assert m1.pf.loc['totcost', 'std_err'] == approx(0.000248, rel=5e-3)


def test_constraint_ordering_1():
	#m0 = larch.Model.Example(1)
	m1 = larch.Model.Example(1)

	# m0.utility_co[2] = P.ASC_SR2 + P('hhinc#23') * X.hhinc
	# m0.utility_co[3] = P.ASC_SR3P + P('hhinc#23') * X.hhinc
	# m0.remove_unused_parameters()

	c3 = OrderingBound("hhinc#3", "hhinc#2")

	# m0.load_data()
	m1.load_data()

	m1.constraints = [c3,]

	# r0 = m0.maximize_loglike(
	# 	method='slsqp',
	# )

	r1 = m1.maximize_loglike(
		method='slsqp',
	)

	assert m1['hhinc#2'].value / m1['hhinc#3'].value == approx(1.0, rel=1e-3)
	assert 'Optimization terminated successfully' in r1.message
	assert c3.fun(m1.pf.value) == approx(0, abs=1e-5)

	# m0.calculate_parameter_covariance()
	m1.calculate_parameter_covariance()

	# assert m0.pf.loc['hhinc#23', 'std_err'] == approx(0.001397, rel=5e-3)
	assert m1.pf.loc['hhinc#2', 'std_err'] == approx(0.001397, rel=5e-3)
	assert m1.pf.loc['hhinc#3', 'std_err'] == approx(0.001397, rel=5e-3)


def test_lower_bound():

	m1 = larch.Model.Example(1)
	m1.set_value('ASC_TRAN', minimum=0)
	m1.load_data()

	# constraint tests are unstable across platforms
	# r1=m1.maximize_loglike(
	# 	method='slsqp',
	# 	options={'ftol': 1e-09},
	# 	quiet=True,
	# )
	# assert 'Optimization terminated successfully' in r1.message
	m1.set_values({'ASC_BIKE': -2.1362348000501323,
				   'ASC_SR2': -2.041879053530548,
				   'ASC_SR3P': -3.568789444368573,
				   'ASC_TRAN': 3.7800630376890545e-12,
				   'ASC_WALK': 0.2099970488220958,
				   'hhinc#2': -0.003442043141464587,
				   'hhinc#3': -0.0010585520966694998,
				   'hhinc#4': -0.012338038801568696,
				   'hhinc#5': -0.014560216659416916,
				   'hhinc#6': -0.012146634316726645,
				   'totcost': -0.00489380761583881,
				   'tottime': -0.059272247433483374})

	m1.calculate_parameter_covariance()

	assert m1.loglike() == approx(-3639.0010819576023)

	assert dict(m1.pf['value']) == approx({
		'ASC_BIKE': -2.136234299750425,
		'ASC_SR2': -2.041878696832068,
		'ASC_SR3P': -3.568790119252386,
		'ASC_TRAN': 0.0,
		'ASC_WALK': 0.20999688402874506,
		'hhinc#2': -0.0034420525424271203,
		'hhinc#3': -0.001058543967530453,
		'hhinc#4': -0.012338043439238819,
		'hhinc#5': -0.014560247040617383,
		'hhinc#6': -0.012146632951225103,
		'totcost': -0.0048938088864863595,
		'tottime': -0.059272225511784224,
	}, abs=1e-10, rel=1e-2)

	se_ = {
		'ASC_BIKE': 0.30102780156769227,
		'ASC_SR2': 0.10147191931629226,
		'ASC_SR3P': 0.17587511190420965,
		'ASC_TRAN': np.nan,
		'ASC_WALK': 0.176809227982435,
		'hhinc#2': 0.0015471476930952455,
		'hhinc#3': 0.002545194004687296,
		'hhinc#4': 0.0012854684301876374,
		'hhinc#5': 0.005332349790044528,
		'hhinc#6': 0.0030443866859315343,
		'totcost': 0.00023967125878600402,
		'tottime': 0.0027703331869058825,
	}
	assert dict(m1.pf['std_err']) == approx(se_, abs=1e-10, rel=1e-2, nan_ok=True)

	assert dict(m1.pf['unconstrained_std_err']) == approx({
		'ASC_BIKE': 0.3048352068287236,
		'ASC_SR2': 0.10486177423525046,
		'ASC_SR3P': 0.17846252159720813,
		'ASC_TRAN': 0.1330469233545672,
		'ASC_WALK': 0.19586113875694,
		'hhinc#2': 0.0015673042154757647,
		'hhinc#3': 0.002560982061812878,
		'hhinc#4': 0.001926809603173801,
		'hhinc#5': 0.005344633810524164,
		'hhinc#6': 0.0030865953487895412,
		'totcost': 0.00023967145587986189,
		'tottime': 0.0032194798668345615,
	}, rel=1e-2)


@pytest.mark.skip(reason="constraint tests are unstable across platforms")
def test_upper_bound():
	#m0 = larch.Model.Example(1)
	m1 = larch.Model.Example(1)

	#m0.lock_value('tottime', -0.1)
	m1.set_value('tottime', maximum=-0.1)

	#m0.load_data()
	m1.load_data()

	# r0 = m0.torch.maximize_loglike(
	# 	method='slsqp',
	# 	options={'ftol': 1e-09},
	# )

	r1 = m1.maximize_loglike(
		method='slsqp',
		options={'ftol': 1e-12},
		quiet=True,
	)

	assert 'Optimization terminated successfully' in r1.message

	#m0.calculate_parameter_covariance()
	m1.calculate_parameter_covariance()

	assert r1.loglike == approx(-3725.072026618359)

	# assert dict(m1.pf['value']) == approx({
	# 	'ASC_BIKE': -1.832639738018494,
	# 	'ASC_SR2': -1.975601900695666,
	# 	'ASC_SR3P': -3.4822938489427426,
	# 	'ASC_TRAN': 0.23349082549768843,
	# 	'ASC_WALK': 1.1111842511724719,
	# 	'hhinc#2': -0.002212546400939385,
	# 	'hhinc#3': 0.00022218881480540761,
	# 	'hhinc#4': -0.0054968611164901825,
	# 	'hhinc#5': -0.012523465955589126,
	# 	'hhinc#6': -0.010794428135692613,
	# 	'totcost': -0.005088026048518891,
	# 	'tottime': -0.1
	# }, rel=1e-2)
	assert m1.pf.loc['tottime', 'value'] == approx(-0.1, rel=1e-3)

	assert dict(m1.pf['std_err']) == approx({
		'ASC_BIKE': 0.30013608145167614,
		'ASC_SR2': 0.10555671986388032,
		'ASC_SR3P': 0.17980894548116735,
		'ASC_TRAN': 0.12312298703458116,
		'ASC_WALK': 0.17807354839871625,
		'hhinc#2': 0.0015763567265900031,
		'hhinc#3': 0.0025749894562580172,
		'hhinc#4': 0.001920571529389678,
		'hhinc#5': 0.005249394394490925,
		'hhinc#6': 0.003146693697448889,
		'totcost': 0.00024654869488240555,
		'tottime': 0
	}, abs=1e-10, rel=1e-2)

	assert dict(m1.pf['unconstrained_std_err']) == approx({
		'ASC_BIKE': 0.30323766594813406,
		'ASC_SR2': 0.1069872808913486,
		'ASC_SR3P': 0.18097129736247558,
		'ASC_TRAN': 0.1439393314870008,
		'ASC_WALK': 0.20788197245193588,
		'hhinc#2': 0.0015763569302413402,
		'hhinc#3': 0.0025750088819569447,
		'hhinc#4': 0.001921038696332169,
		'hhinc#5': 0.005249394449971851,
		'hhinc#6': 0.0031492030054873824,
		'totcost': 0.0002472663647304817,
		'tottime': 0.004377145909550393,
	}, rel=1e-2)

@pytest.mark.skip(reason="constraint tests are unstable across platforms")
def test_multi_constraints():
	#m0 = larch.Model.Example(1)
	m1 = larch.Model.Example(1)

	#m0.lock_value('tottime', -0.1)
	m1.set_value('tottime', maximum=-0.1)

	#m0.utility_co[2] = P.ASC_SR2 + P('hhinc#23') * X.hhinc
	#m0.utility_co[3] = P.ASC_SR3P + P('hhinc#23') * X.hhinc
	#m0.remove_unused_parameters()

	m1.constraints.append(OrderingBound("hhinc#3 <= hhinc#2"))

	#m0.load_data()
	m1.load_data()

	# r0 = m0.torch.maximize_loglike(
	# 	method='slsqp',
	# 	options={'ftol': 1e-09},
	# )

	r1 = m1.maximize_loglike(
		method='slsqp',
		options={'ftol': 1e-09},
		quiet=True,
	)
	# assert r1.message == 'Positive directional derivative for linesearch'
	m1.constraints.rescale(0.1)

	r1 = m1.maximize_loglike(
		method='slsqp',
		options={'ftol': 1e-09},
		quiet=True,
	)
	# assert r1.message == 'Positive directional derivative for linesearch'

	m1.constraints.rescale(0.01)
	r1 = m1.maximize_loglike(
		method='slsqp',
		options={'ftol': 1e-09},
		quiet=True,
	)

	assert r1.loglike == approx(-3725.439832484451)

	assert 'Optimization terminated successfully' in r1.message

	#m0.calculate_parameter_covariance()

	m1.calculate_parameter_covariance()

	assert dict(m1.pf['value']) == approx({
		'ASC_BIKE': -1.8336226490217296,
		'ASC_SR2': -2.008575006161787,
		'ASC_SR3P': -3.372494233637122,
		'ASC_TRAN': 0.23543257975862672,
		'ASC_WALK': 1.1106100812079671,
		'hhinc#2': -0.0016453210270738865,
		'hhinc#3': -0.001645187376205547,
		'hhinc#4': -0.0055451755658633,
		'hhinc#5': -0.012519484434344206,
		'hhinc#6': -0.01079425276590098,
		'totcost': -0.005093387068047046,
		'tottime': -0.1
	}, rel=1e-2)

	# Problem: Travis is failing on this test, giving
	#    tottime as NaN.  Disabling until we find a
	#    similar problem we can test and diagnose
	assert dict(m1.pf['std_err']) == approx({
		'ASC_BIKE': 0.30013699696981455,
		'ASC_SR2': 0.09856612504524623,
		'ASC_SR3P': 0.1250682262709841,
		'ASC_TRAN': 0.12311806485021345,
		'ASC_WALK': 0.17807446535212992,
		'hhinc#2': 0.0014236497001240312,
		'hhinc#3': 0.001423649700124031,
		'hhinc#4': 0.0019212691279916797,
		'hhinc#5': 0.005248506439174827,
		'hhinc#6': 0.003146916477051659,
		'totcost': 0.0002466030726673273,
		'tottime': 0
	}, abs=1e-10, rel=5e-2)

	assert dict(m1.pf['unconstrained_std_err']) == approx({
		'ASC_BIKE': 0.3032386370524189,
		'ASC_SR2': 0.10695557451038115,
		'ASC_SR3P': 0.18033743391454088,
		'ASC_TRAN': 0.1439964957540911,
		'ASC_WALK': 0.20789603304640025,
		'hhinc#2': 0.0015638823624232918,
		'hhinc#3': 0.002640071485204856,
		'hhinc#4': 0.00192260391409179,
		'hhinc#5': 0.0052485070222418155,
		'hhinc#6': 0.0031494462988719317,
		'totcost': 0.00024742269603570923,
		'tottime': 0.004377029904009681
	}, rel=5e-2)

def test_overspec():

	m0 = larch.Model.Example(1)
	m0.utility_ca = m0.utility_ca + P.failpar * X('1')
	m0.utility_co[1] = P.ASC_DA
	m0.lock_value('tottime', -0.1)
	m0.utility_co[2] = P.ASC_SR2 + P('hhinc#23') * X.hhinc
	m0.utility_co[3] = P.ASC_SR3P + P('hhinc#23') * X.hhinc
	m0.remove_unused_parameters()
	m0.load_data()
	# constraint tests are unstable across platforms
	# r0 = m0.maximize_loglike(
	# 	quiet=True,
	# )
	m0.set_values({
		'ASC_BIKE': -0.8550063261138748,
		'ASC_DA': 0.9780172816142935,
		'ASC_SR2': -1.0303087193826583,
		'ASC_SR3P': -2.394702207497934,
		'ASC_TRAN': 1.2134607482035888,
		'ASC_WALK': 2.0885392231767055,
		'failpar': -1.2138930556454395e-14,
		'hhinc#23': -0.001647452425832848,
		'hhinc#4': -0.005545798283823439,
		'hhinc#5': -0.012530050562373019,
		'hhinc#6': -0.010792561322141715,
		'totcost': -0.005093524162084949,
		'tottime': -0.1,
	})
	m0.calculate_parameter_covariance()
	possover = m0.possible_overspecification
	assert possover.data.shape == (7, 2)
	assert all(
		possover.data.index.sort_values() == [
			'ASC_BIKE', 'ASC_DA', 'ASC_SR2', 'ASC_SR3P',
			'ASC_TRAN', 'ASC_WALK', 'failpar',
		])

def test_parameter_summary():
	import larch
	from larch.model.constraints import RatioBound, OrderingBound, FixedBound

	m0 = larch.Model.Example(1)
	m0.lock_value('tottime', -0.05)
	m0.set_value('totcost', maximum=-0.005, minimum=-0.02)
	m0.constraints.append(OrderingBound("hhinc#3 <= hhinc#2"))
	m0.load_data()

	# constraint tests are unstable across platforms
	# r0 = m0.maximize_loglike(
	# 	method='slsqp',
	# 	options={'ftol': 1e-09},
	# 	quiet=True,
	# )
	m0.set_values({
		'ASC_BIKE': -2.401574701017008,
		'ASC_SR2': -2.2234574452767037,
		'ASC_SR3P': -3.630001719635557,
		'ASC_TRAN': -0.7000998451196682,
		'ASC_WALK': -0.25495834390295924,
		'hhinc#2': -0.0015948074747861667,
		'hhinc#3': -0.0015948074747729564,
		'hhinc#4': -0.005385664916039662,
		'hhinc#5': -0.01284021327300144,
		'hhinc#6': -0.009658848920781143,
		'totcost': -0.005000000000003634,
		'tottime': -0.05,
	})

	m0.calculate_parameter_covariance()

	ps = m0.parameter_summary('df').data
	stable_df(ps, 'parameter_summary_test')

	try:
		assert ps.loc[('LOS', 'totcost'), 'Value'] == "-0.00500"
		assert ps.loc[('LOS', 'tottime'), 'Value'] == "-0.0500"
		assert ps.loc[('Income', 'hhinc#2'), 'Value'] == "-0.00159"
		assert ps.loc[('Income', 'hhinc#3'), 'Value'] == "-0.00159"

		assert ps.loc[('LOS', 'totcost'), 'Std Err'] == " NA"
		assert ps.loc[('LOS', 'tottime'), 'Std Err'] == " NA"
		assert ps.loc[('Income', 'hhinc#2'), 'Std Err'] == " 0.00140"
		assert ps.loc[('Income', 'hhinc#3'), 'Std Err'] == " 0.00140"

		assert ps.loc[('LOS', 'totcost'), 't Stat'] == " NA"
		assert ps.loc[('LOS', 'tottime'), 't Stat'] == " NA"
		assert ps.loc[('Income', 'hhinc#2'), 't Stat'] == "-1.14"
		assert ps.loc[('Income', 'hhinc#3'), 't Stat'] == "-1.14"

		assert ps.loc[('LOS', 'totcost'), 'Signif'] == "[***]"
		assert ps.loc[('LOS', 'tottime'), 'Signif'] == ""
		assert ps.loc[('Income', 'hhinc#2'), 'Signif'] == ""
		assert ps.loc[('Income', 'hhinc#3'), 'Signif'] == ""

		assert ps.loc[('LOS', 'totcost'), 'Constrained'] == "totcost ≤ -0.005"
		assert ps.loc[('LOS', 'tottime'), 'Constrained'] == "fixed value"
		assert ps.loc[('Income', 'hhinc#2'), 'Constrained'] == "hhinc#3 ≤ hhinc#2"
		assert ps.loc[('Income', 'hhinc#3'), 'Constrained'] == "hhinc#3 ≤ hhinc#2"
	except:
		print(ps.iloc[:,:3])
		print(ps.iloc[:,3:])
		raise

def test_contraint_interpretation():

	assert interpret_contraint("aaaa > bbbb") == OrderingBound('bbbb', 'aaaa')
	assert interpret_contraint("aaaa < bbbb") == OrderingBound('aaaa', 'bbbb')
	assert interpret_contraint("aaaa/bbbb > 3") == RatioBound('aaaa', 'bbbb', min_ratio=3)
	assert interpret_contraint("aaaa/bbbb < 3") == RatioBound('aaaa', 'bbbb', max_ratio=3)
	assert interpret_contraint("aaaa / bbbb > 3") == RatioBound('aaaa', 'bbbb', min_ratio=3)
	assert interpret_contraint("aaaa / bbbb < 3") == RatioBound('aaaa', 'bbbb', max_ratio=3)
	assert interpret_contraint("aaaa / bbbb >= 3") == RatioBound('aaaa', 'bbbb', min_ratio=3)
	assert interpret_contraint("aaaa / bbbb <= 3") == RatioBound('aaaa', 'bbbb', max_ratio=3)
	assert interpret_contraint("3.1 < aaaa / bbbb <= 3.2") == RatioBound('aaaa', 'bbbb', min_ratio=3.1, max_ratio=3.2)
	assert interpret_contraint("aaaa > 3") == FixedBound('aaaa', minimum=3)
	assert interpret_contraint("aaaa < 3") == FixedBound('aaaa', maximum=3)
	assert interpret_contraint("3.1 < aaaa < 3.5") == FixedBound('aaaa', minimum=3.1, maximum=3.5)

	assert interpret_contraint("aaaa/bbbb > 1/2") == RatioBound('aaaa', 'bbbb', min_ratio=1/2)
	assert interpret_contraint("aaaa/bbbb < 1/2") == RatioBound('aaaa', 'bbbb', max_ratio=1/2)
	assert interpret_contraint("aaaa / bbbb > 1/2") == RatioBound('aaaa', 'bbbb', min_ratio=1/2)
	assert interpret_contraint("aaaa / bbbb < 1/2") == RatioBound('aaaa', 'bbbb', max_ratio=1/2)
	assert interpret_contraint("aaaa / bbbb >= 1/2") == RatioBound('aaaa', 'bbbb', min_ratio=1/2)
	assert interpret_contraint("aaaa / bbbb <= 1/2") == RatioBound('aaaa', 'bbbb', max_ratio=1/2)
	assert interpret_contraint("1/4 < aaaa / bbbb <= 3/4") == RatioBound('aaaa', 'bbbb', min_ratio=1/4, max_ratio=3/4)
	assert interpret_contraint("aaaa > 3/4") == FixedBound('aaaa', minimum=3/4)
	assert interpret_contraint("aaaa < 3/4") == FixedBound('aaaa', maximum=3/4)
	assert interpret_contraint("1/4 < aaaa < 3/4") == FixedBound('aaaa', minimum=1/4, maximum=3/4)

def test_constraint_parameter_summary_ratios():

	m = larch.Model.Example(1)
	m.set_value('totcost', minimum=-0.001)
	m.set_value('hhinc#3', minimum=0.0025)
	m.set_value('hhinc#5', minimum=-0.008)
	m.lock_value('hhinc#6', value=-0.009)
	m.load_data()
	m.set_values({
		'ASC_BIKE': -2.203264336154524,
		'ASC_SR2': -1.8913348121897215,
		'ASC_SR3P': -3.2288258012739846,
		'ASC_TRAN': -0.520325914514216,
		'ASC_WALK': 0.12488474492871472,
		'hhinc#2': -0.0010437239183118503,
		'hhinc#3': 0.002500000762078847,
		'hhinc#4': -0.0017845440559583583,
		'hhinc#5': -0.007999992934338814,
		'hhinc#6': -0.009,
		'totcost': -0.0009999993257563983,
		'tottime': -0.052413890729296815,
	})
	m.calculate_parameter_covariance()
	stable_df(m.parameter_summary().data, 'like_ratio_test')



## Exampville

from ..data_warehouse import example_file

# flog = larch.logging.flogger(level=30, label="exampville")
def flog(a0, *arg, **kwarg):
	pass


__all__ = ['files', ]

_cache_1 = {}

_directory = None


def build_directory(directory=None):
	global _directory
	if directory is None:
		if _directory is not None:
			directory = _directory
		else:
			from ..util.temporaryfile import TemporaryDirectory
			_directory = directory = TemporaryDirectory()
	else:
		_directory = directory
	return _directory

class _files:

	def __init__(self, tag=None):
		if tag is None or tag == 0:
			self._tag = ''
		else:
			self._tag = f'_{tag}'

	@property
	def shapefile(self):
		return example_file(f'exampville_taz{self._tag}.zip', rel=True)

	@property
	def employment(self):
		return example_file(f'exampville_employment{self._tag}.csv.gz', missing_ok=True, rel=True)

	@property
	def hh(self):
		return example_file(f'exampville_households{self._tag}.csv.gz', missing_ok=True, rel=True)

	@property
	def person(self):
		return example_file(f'exampville_persons{self._tag}.csv.gz', missing_ok=True, rel=True)

	@property
	def tour(self):
		return example_file(f'exampville_tours{self._tag}.csv.gz', missing_ok=True, rel=True)

	@property
	def skims(self):
		return example_file(f'exampville_skims{self._tag}.omx', missing_ok=True, rel=True)

	@property
	def demographics(self):
		return example_file(f'exampville_demographics{self._tag}.csv.gz', missing_ok=True, rel=True)



files = _files()
files_alt = _files(42)

# The builder is abandoned in favor of some fixed pre-built files.
#
# def build_year_1(
# 		n_HH=5000,
# 		directory=None,
# 		seed=0,
# 		output_format='csv',
# ):
#
# 	global _cache_1, _directory
# 	if output_format in _cache_1:
# 		return _cache_1[output_format]
#
# 	import geopandas
# 	taz_shape = geopandas.read_file("zip://" + files.shapefile)
#
# 	nZones = len(taz_shape)
#
# 	flog("EXAMPVILLE Builder (Year 1)")
# 	flog("  simulating a survey of {} households", n_HH)
# 	flog("  traveling among {} travel analysis zones", nZones)
#
# 	if directory is None:
# 		if _directory is not None:
# 			directory = _directory
# 		else:
# 			from ..util.temporaryfile import TemporaryDirectory
# 			_directory = directory = TemporaryDirectory()
# 	else:
# 		_directory = directory
#
# 	if isinstance(transit_scope, tuple):
# 		transit_scope = slice(*transit_scope)
#
# 	if transit_scope.stop > nZones:
# 		raise TypeError('transit_scope too large for nZones')
#
# 	# The randomizer seed is reset to zero by default so that we get consistent generated
# 	# Exampville survey results.  (High levels of randomness are not very important
# 	# here because we just want to demonstrate the models.)
# 	numpy.random.seed(seed)
#
# 	## Zones
# 	flog("Zones")
# 	pop_weight = numpy.fmax(numpy.arange(nZones), numpy.flipud(numpy.arange(nZones))).astype(float)
# 	pop_weight /= pop_weight.sum()
# 	wrk_weight = scipy.stats.binom.pmf(numpy.arange(nZones), nZones - 1, .5)
# 	wrk_weight /= wrk_weight.sum()
#
# 	zone_lat = (-1) ** numpy.arange(nZones).astype(numpy.int64)
# 	zone_lon = 11 + numpy.arange(nZones).astype(numpy.int64)
#
# 	## Skims
# 	flog("Skims")
# 	distance = numpy.zeros([nZones, nZones])
# 	gaps = numpy.random.random(nZones) + 1
# 	for z in range(nZones):
# 		for y in range(z, nZones):
# 			distance[y, z] = distance[z, y] = gaps[z:y].sum()
# 		distance[z, z] = numpy.random.random()
#
# 	drivetime = (numpy.random.exponential(1, [nZones, nZones]) + 1) * distance
# 	drivetime[drivetime < 2.0] = 2.0
# 	hov_time = drivetime.copy()
# 	hov_time[:nZones // 2, nZones // 2:] -= numpy.random.random() * 1.5 + 0.5
# 	hov_time[nZones // 2:, :nZones // 2] -= numpy.random.random() * 1.5 + 0.5
#
# 	transit_range = transit_scope.stop - transit_scope.start
# 	transittime = numpy.zeros([nZones, nZones])
# 	transittime[transit_scope, transit_scope] = (numpy.random.random([transit_range, transit_range]) + 1) * distance[
# 		transit_scope, transit_scope]
#
# 	transitfare = numpy.zeros([nZones, nZones])
# 	transitfare[transit_scope, transit_scope] = 1.5
#
# 	## Households
# 	flog("HHs")
# 	HHidx = numpy.arange(n_HH, dtype=numpy.int64)
# 	HHid = numpy.asarray([50000 + i for i in HHidx])
# 	HHincome = numpy.round(numpy.random.normal(75000, 25000, [n_HH, ]), -3).astype(numpy.int64)
# 	HHsize = numpy.floor(numpy.random.exponential(0.8, [n_HH, ]) + 1 + numpy.random.random([n_HH, ])).astype(
# 		numpy.int64)
# 	HHhomezone = numpy.random.choice(numpy.arange(1, nZones + 1), size=[n_HH, ], replace=True, p=pop_weight).astype(numpy.int64)
#
# 	## People
# 	flog("People")
# 	n_PER = numpy.sum(HHsize)
# 	PERidx = numpy.arange(n_PER, dtype=numpy.int64)
# 	PERid = numpy.asarray([60000 + i for i in PERidx])
# 	PERhhid = numpy.zeros(n_PER, dtype=numpy.int64)
# 	PERhhidx = numpy.zeros(n_PER, dtype=numpy.int64)
# 	n2 = 0
# 	for n1 in range(n_HH):
# 		PERhhid[n2:(n2 + HHsize[n1])] = HHid[n1]
# 		PERhhidx[n2:(n2 + HHsize[n1])] = HHidx[n1]
# 		n2 += HHsize[n1]
# 	PERage = (numpy.random.random(n_PER) * 80 + 5).astype(numpy.int64)
# 	PERworks = ((numpy.random.random(n_PER) > 0.2) & (PERage > 16) & (PERage < 70)).astype(numpy.int64)
#
# 	zone_employment = numpy.round(PERworks.sum() * wrk_weight, 0) + 1
#
# 	total_employment = zone_employment.sum()
# 	mean_employment = total_employment / nZones
# 	zone_retail = numpy.fmin(numpy.round(numpy.random.random(nZones) * mean_employment, 0), zone_employment)
# 	zone_nonretail = zone_employment - zone_retail
#
# 	PERnworktours = numpy.random.choice([0, 1, 2, 3], size=[n_PER, ], replace=True, p=[0.1, 0.8, 0.07, 0.03]).astype(
# 		numpy.int64) * PERworks
# 	PERnothertours = numpy.random.choice([0, 1, 2, 3], size=[n_PER, ], replace=True, p=[0.2, 0.5, 0.2, 0.1]).astype(numpy.int64)
# 	PERntours = PERnworktours + PERnothertours
#
# 	## Tours
# 	flog("Tours")
#
# 	n_TOUR = PERntours.sum()
#
# 	TOURid = numpy.arange(n_TOUR, dtype=numpy.int64)
# 	TOURper = numpy.zeros(n_TOUR, dtype=numpy.int64)
# 	TOURperidx = numpy.zeros(n_TOUR, dtype=numpy.int64)
# 	TOURhh = numpy.zeros(n_TOUR, dtype=numpy.int64)
# 	TOURhhidx = numpy.zeros(n_TOUR, dtype=numpy.int64)
# 	TOURdtaz = numpy.zeros(n_TOUR, dtype=numpy.int64)
# 	TOURmode = numpy.zeros(n_TOUR, dtype=numpy.int64)
# 	TOURpurpose = numpy.zeros(n_TOUR, dtype=numpy.int64)
#
# 	# Work tours, then other tours
# 	n2 = 0
# 	for n1 in range(n_PER):
# 		TOURper[n2:(n2 + PERntours[n1])] = PERid[n1]
# 		TOURperidx[n2:(n2 + PERntours[n1])] = PERidx[n1]
# 		TOURhh[n2:(n2 + PERntours[n1])] = PERhhid[n1]
# 		TOURhhidx[n2:(n2 + PERntours[n1])] = PERhhidx[n1]
# 		TOURpurpose[n2:(n2 + PERnworktours[n1])] = 1
# 		TOURpurpose[(n2 + PERnworktours[n1]):(n2 + PERntours[n1])] = 2
# 		n2 += PERntours[n1]
#
# 	#### Utility by mode to various destinations
# 	flog("Choice Probability")
# 	nameModes = ['DA', 'SR', 'Walk', 'Bike', 'Transit']
# 	mDA = 0
# 	mSR = 1
# 	mWA = 2
# 	mBI = 3
# 	mTR = 4
# 	nModes = len(nameModes)
#
# 	nModeNests = 3
#
# 	paramCOST = -0.312
# 	paramTIME = -0.123
# 	paramNMTIME = -0.246
# 	paramDIST = -0.00357
# 	paramLNDIST = -0.00642
#
# 	paramMUcar = 0.5
# 	paramMUnon = 0.75
# 	paramMUmot = 0.8
# 	paramMUtop = 1.0
#
# 	Util = numpy.zeros([n_TOUR, nZones, nModes])
# 	for n in range(n_TOUR):
# 		# Mode
# 		#		flog('N {}',n)
# 		#		flog('Util[n,:,:]  &&&')
# 		#		flog('{}',Util[n,:,:])
# 		otazi = HHhomezone[TOURhhidx[n]] - 1
# 		Util[n, :, mDA] += drivetime[otazi, :] * paramTIME + distance[otazi, :] * 0.20 * paramCOST
# 		if HHincome[TOURhhidx[n]] >= 75000:
# 			Util[n, :, mDA] += 1.0
# 			Util[n, :, mTR] -= 0.5
# 		Util[n, :, mSR] += drivetime[otazi, :] * paramTIME - 1.0 + distance[otazi, :] * 0.20 * 0.5 * paramCOST
# 		Util[n, :, mWA] += 1.0 + distance[otazi, :] / 2.5 * 60 * paramNMTIME
# 		Util[n, :, mBI] += -1.25 + distance[otazi, :] / 12 * 60 * paramNMTIME
# 		Util[n, :, mTR] += -1.5 + transittime[otazi, :] * paramTIME + transitfare[otazi, :] * paramCOST
# 		# Destination
# 		Util[n, :, :] += distance[otazi, :, None] * paramDIST + log1p(distance[otazi, :, None]) * paramLNDIST
# 		if HHincome[TOURhhidx[n]] <= 50000:
# 			Util[n, :, :] += 0.75 * log(zone_retail * 2.71828 + zone_nonretail)[:, None]
# 		else:
# 			Util[n, :, :] += 0.75 * log(zone_retail + zone_nonretail * 2.71828)[:, None]
# 		# flog('Util[n,:,:]  ...')
# 		#		flog('{}',Util[n,:,:])
# 		# Unavails
# 		if PERage[TOURperidx[n]] < 16:
# 			Util[n, :, mDA] = -numpy.inf
# 		Util[n, transitfare[otazi, :] <= 0, mTR] = -numpy.inf
# 		Util[n, distance[otazi, :] >= 3, mWA] = -numpy.inf
# 		Util[n, distance[otazi, :] >= 15, mBI] = -numpy.inf
# 	# flog('Util[n,:,:]  +++')
# 	#		flog('{}',Util[n,:,:])
#
#
# 	CPr_car = numpy.zeros([n_TOUR, nZones, 2])  # [DA,SR]
# 	CPr_non = numpy.zeros([n_TOUR, nZones, 2])  # [WA,BI]
# 	CPr_mot = numpy.zeros([n_TOUR, nZones, 2])  # [TR,Car]
# 	CPr_top = numpy.zeros([n_TOUR, nZones, 2])  # [Non,Mot]
#
# 	NLS_car = numpy.zeros([n_TOUR, nZones, ])
# 	NLS_non = numpy.zeros([n_TOUR, nZones, ])
# 	NLS_mot = numpy.zeros([n_TOUR, nZones, ])
# 	MLS_top = numpy.zeros([n_TOUR, nZones, ])  # Mode choice logsum
# 	DLS_top = numpy.zeros([n_TOUR, ])  # Dest choice logsum
#
# 	Pr_modes = numpy.zeros([n_TOUR, nZones, nModes])
# 	Pr_dest = numpy.zeros([n_TOUR, nZones])
#
# 	with numpy.errstate(divide='ignore', invalid='ignore'):
# 		for n in range(n_TOUR):
# 			NLS_car[n, :] = paramMUcar * log(exp(Util[n, :, mDA] / paramMUcar) + exp(Util[n, :, mSR] / paramMUcar))
# 			NLS_non[n, :] = paramMUnon * log(exp(Util[n, :, mWA] / paramMUnon) + exp(Util[n, :, mBI] / paramMUnon))
# 			NLS_mot[n, :] = paramMUmot * log(exp(NLS_car[n, :] / paramMUmot) + exp(Util[n, :, mTR] / paramMUmot))
# 			MLS_top[n, :] = log(exp(NLS_non[n, :]) + exp(NLS_mot[n, :]))
# 			DLS_top[n] = log(numpy.sum(exp(MLS_top[n, :])))
#
# 			Pr_dest[n, :] = exp(MLS_top[n, :] - DLS_top[n])
#
# 			CPr_top[n, :, 0] = exp((NLS_non[n, :] - MLS_top[n, :]) / paramMUtop)
# 			CPr_top[n, :, 1] = exp((NLS_mot[n, :] - MLS_top[n, :]) / paramMUtop)
# 			CPr_mot[n, :, 0] = exp((Util[n, :, mTR] - NLS_mot[n, :]) / paramMUmot)
# 			CPr_mot[n, :, 1] = exp((NLS_car[n, :] - NLS_mot[n, :]) / paramMUmot)
# 			CPr_non[n, :, 0] = exp((Util[n, :, mWA] - NLS_non[n, :]) / paramMUnon)
# 			CPr_non[n, :, 1] = exp((Util[n, :, mBI] - NLS_non[n, :]) / paramMUnon)
# 			CPr_car[n, :, 0] = exp((Util[n, :, mDA] - NLS_car[n, :]) / paramMUcar)
# 			CPr_car[n, :, 1] = exp((Util[n, :, mSR] - NLS_car[n, :]) / paramMUcar)
#
# 			Pr_modes[n, :, mTR] = CPr_mot[n, :, 0] * CPr_top[n, :, 1] * Pr_dest[n, :]
# 			Pr_modes[n, :, mWA] = CPr_non[n, :, 0] * CPr_top[n, :, 0] * Pr_dest[n, :]
# 			Pr_modes[n, :, mBI] = CPr_non[n, :, 1] * CPr_top[n, :, 0] * Pr_dest[n, :]
# 			Pr_modes[n, :, mDA] = CPr_car[n, :, 0] * CPr_mot[n, :, 1] * CPr_top[n, :, 1] * Pr_dest[n, :]
# 			Pr_modes[n, :, mSR] = CPr_car[n, :, 1] * CPr_mot[n, :, 1] * CPr_top[n, :, 1] * Pr_dest[n, :]
#
# 	Pr_modes[numpy.isnan(Pr_modes)] = 0
#
# 	## Choices
# 	flog("Choices")
# 	for n in range(n_TOUR):
# 		try:
# 			ch = numpy.random.choice(nModes * nZones, replace=True, p=Pr_modes[n, :, :].ravel())
# 		except:
# 			flog("total prob = {}", Pr_modes[n, :, :].sum())
# 			raise
# 		dtazi = ch // nModes
# 		modei = ch - (dtazi * nModes)
# 		TOURdtaz[n] = dtazi + 1
# 		TOURmode[n] = modei + 1
#
# 	### Write Out Data
# 	flog("Output")
#
# 	if not os.path.exists(directory):
# 		os.makedirs(directory)
#
# 	from ..omx import OMX
#
# 	omx = OMX(os.path.join(directory, 'exampville.omx'), mode='a')
# 	omx.shape = (nZones, nZones)
# 	omx.add_matrix('DIST', distance)
# 	omx.add_matrix('AUTO_TIME', drivetime)
# 	omx.add_matrix('RAIL_TIME', transittime)
# 	omx.add_matrix('RAIL_FARE', transitfare)
# 	omx.add_lookup('TAZID', numpy.arange(1, nZones + 1, dtype=numpy.int64))
# 	omx.add_lookup('EMPLOYMENT', zone_employment)
# 	omx.add_lookup('EMP_RETAIL', zone_retail)
# 	omx.add_lookup('EMP_NONRETAIL', zone_nonretail)
# 	omx.add_lookup('LAT', zone_lat)
# 	omx.add_lookup('LON', zone_lon)
#
# 	omx.flush()
# 	omx.close()
# 	omx = OMX(os.path.join(directory, 'exampville.omx'), mode='r')
#
# 	if output_format == 'h5':
#
# 		from ..data_services.h5 import H5Pod
#
# 		f_hh = H5Pod(os.path.join(directory, 'exampville_hh.h5'), mode='a')
# 		f_hh.add_array('HHID', HHid)
# 		f_hh.add_array('INCOME', HHincome)
# 		f_hh.add_array('HHSIZE', HHsize)
# 		f_hh.add_array('HOMETAZ', HHhomezone)
# 		f_hh.flush()
# 		f_hh_filename = f_hh.filename
#
# 		f_pp = H5Pod(os.path.join(directory, 'exampville_person.h5'), mode='a')
# 		f_pp.add_array('PERSONID', PERid)
# 		f_pp.add_array('HHID', PERhhid)
# 		f_pp.add_array('AGE', PERage)
# 		f_pp.add_array('WORKS', PERworks, dictionary={1: 'Yes', 0: 'No'}, title='Person has a regular job')
# 		f_pp.add_array('N_WORKTOURS', PERnworktours,
# 					   title='Number of work tours reported by this person on the survey day')
# 		f_pp.add_array('N_OTHERTOURS', PERnothertours,
# 					   title='Number of non-work tours reported by this person on the survey day')
# 		f_pp.add_array('N_TOTALTOURS', PERntours,
# 					   title='Number of non-work tours reported by this person on the survey day')
# 		f_pp.flush()
# 		f_pp_filename = f_pp.filename
#
# 		f_tour = H5Pod(os.path.join(directory, 'exampville_tours.h5'), mode='a')
# 		f_tour.add_array('TOURID', TOURid)
# 		f_tour.add_array('HHID', TOURhh)
# 		f_tour.add_array('PERSONID', TOURper)
# 		f_tour.add_array('DTAZ', TOURdtaz)
# 		f_tour.add_array('TOURMODE', TOURmode, dictionary={
# 			1: 'DA',
# 			2: 'SR',
# 			3: 'Walk',
# 			4: 'Bike',
# 			5: 'Transit',
# 		})
# 		f_tour.add_array('TOURPURP', TOURpurpose, dictionary={
# 			1: 'Work Tour',
# 			2: 'Non-Work Tour',
# 		})
# 		f_tour.flush()
# 		f_tour_filename = f_tour.filename
#
# 	elif output_format == 'csv':
#
# 		import pandas
# 		from collections import OrderedDict
# 		f_hh = pandas.DataFrame.from_dict(
# 			OrderedDict([
# 				('HHID', HHid),
# 				('INCOME', HHincome),
# 				('HHSIZE', HHsize),
# 				('HOMETAZ', HHhomezone),
# 			])
# 		)
# 		f_hh_filename = os.path.join(directory, 'exampville_hh.csv')
# 		f_hh.to_csv(f_hh_filename)
#
# 		f_pp = pandas.DataFrame.from_dict(
# 			OrderedDict([
# 				('PERSONID', PERid),
# 				('HHID', PERhhid),
# 				('AGE', PERage),
# 				('WORKS', PERworks),
# 				('N_WORKTOURS', PERnworktours),
# 				('N_OTHERTOURS', PERnothertours),
# 				('N_TOTALTOURS', PERntours),
# 			])
# 		)
# 		f_pp_filename = os.path.join(directory, 'exampville_person.csv')
# 		f_pp.to_csv(f_pp_filename)
#
# 		f_tour = pandas.DataFrame.from_dict(
# 			OrderedDict([
# 				('TOURID', TOURid),
# 				('HHID', TOURhh),
# 				('PERSONID', TOURper),
# 				('DTAZ', TOURdtaz),
# 				('TOURMODE', TOURmode),
# 				('TOURPURP', TOURpurpose),
# 			])
# 		)
# 		f_tour_filename = os.path.join(directory, 'exampville_tours.csv')
# 		f_tour.to_csv(f_tour_filename)
#
# 	else:
# 		raise ValueError(f'bad output_format "{output_format}"')
#
# 	flog("EXAMPVILLE Completed Builder (Year 1)")
# 	flog("   SKIMS  : {}", omx.filename)
# 	flog("   HHs    : {}", f_hh_filename)
# 	flog("   Persons: {}", f_pp_filename)
# 	flog("   Tours  : {}", f_tour_filename)
#
# 	_cache_1[output_format] = (directory, omx, f_hh, f_pp, f_tour)
#
# 	if output_format == 'h5':
# 		return directory, omx, f_hh.astype('idco'), f_pp.astype('idco'), f_tour.astype('idco')
# 	else:
# 		return directory, omx, f_hh, f_pp, f_tour
#
# builder = builder_1 = build_year_1
#



import numpy as np
import pandas as pd
import re
import itertools
from typing import Mapping
from larch import P, X, DataFrames, Model
from larch.model.abstract_model import AbstractChoiceModel
from larch.model.tree import NestingTree
from larch.util import Dict

import logging
from ...log import logger_name

_logger = logging.getLogger(logger_name)


def cv_to_ca(alt_values, dtype="float64"):
    """
    Convert a choosers-variables DataFrame to an idca DataFrame.

    Parameters
    ----------
    alt_values : pandas.DataFrame
        This DataFrame should be in choosers-variables format,
        with one row per chooser and variable, and one column per
        alternative.  The id's for the choosers and variables
        must be in that order, in a two-level MultiIndex.
    dtype : dtype
        Convert the incoming data to this type.  Set to None to
        skip data conversion.

    Returns
    -------
    pandas.DataFrame
        The resulting DataFrame is transformed into Larch's
        idca format, with one row per chooser (case) and
        alternative, and one column per variable.
    """

    # Read the source file, converting to a tall stack of
    # data with a 3 level multiindex
    x_ca_tall = alt_values.stack()

    # Set the last level index name to 'altid'
    x_ca_tall.index.rename("altid", -1, inplace=True)
    c_, v_, a_ = x_ca_tall.index.names

    # case and alt id's should be integers.
    # Easier to manipulate dtypes of a multiindex if we just convert
    # it to columns first.  If this ends up being a performance issue
    # we can look at optimizing this in the future.
    x_ca_tall = x_ca_tall.reset_index()
    x_ca_tall[c_] = x_ca_tall[c_].astype(int)
    x_ca_tall[a_] = x_ca_tall[a_].astype(int)
    x_ca_tall = x_ca_tall.set_index([c_, v_, a_])

    if dtype is not None:
        # Convert instances where data is 'False' or 'True' to numbers
        x_ca_tall[x_ca_tall == "False"] = 0
        x_ca_tall[x_ca_tall == "True"] = 1

        # Convert data to float64 to optimize computation speed in larch
        x_ca_tall = x_ca_tall.astype(dtype)

    # Unstack the variables dimension
    x_ca = x_ca_tall.unstack(1)

    # Code above added a dummy top level to columns, remove it here.
    x_ca.columns = x_ca.columns.droplevel(0)

    return x_ca


def prevent_overlapping_column_names(x_ca, x_co):
    """
    Rename columns in idca data to prevent overlapping names.

    Parameters
    ----------
    x_ca, x_co : pandas.DataFrame
        The idca and idco data, respectively

    Returns
    -------
    x_ca, x_co
    """
    renaming = {i: f"{i}_ca" for i in x_ca.columns if i in x_co.columns}
    x_ca.rename(columns=renaming, inplace=True)
    return x_ca, x_co


def linear_utility_from_spec(spec, x_col, p_col, ignore_x=(), segment_id=None):
    """
    Create a linear function from a spec DataFrame.

    Parameters
    ----------
    spec : pandas.DataFrame
        A spec for an ActivitySim model.
    x_col: str
        The name of the columns in spec representing the data.
    p_col: str or dict
        The name of the columns in spec representing the parameters.
        Give as a string for a single column, or as a dict to
        have segments on multiple columns. If given as a dict,
        the keys give the names of the columns to use, and the
        values give the identifiers that will need to match the
        loaded `segment_id` value.
    ignore_x : Collection, optional
        Labels in the spec file to ignore.  Typically this
        includes variables that are pre-processed by ActivitySim
        and therefore don't need to be made available in Larch.
    segment_id : str, optional
        The CHOOSER_SEGMENT_COLUMN_NAME identified for ActivitySim.
        This value is ignored if `p_col` is a string, and required
        if `p_col` is a dict.

    Returns
    -------
    LinearFunction_C
    """
    if isinstance(p_col, dict):
        if segment_id is None:
            raise ValueError("segment_id must be given if p_col is a dict")
        partial_utility = {}
        for seg_p_col, segval in p_col.items():
            partial_utility[seg_p_col] = linear_utility_from_spec(
                spec, x_col, seg_p_col, ignore_x,
            ) * X(f"{segment_id}=={segval}")
        return sum(partial_utility.values())
    return sum(
        P(getattr(i, p_col)) * X(getattr(i, x_col))
        for i in spec.itertuples()
        if (getattr(i, x_col) not in ignore_x) and not pd.isna(getattr(i, p_col))
    )


def dict_of_linear_utility_from_spec(spec, x_col, p_col, ignore_x=()):
    """
    Create a linear function from a spec DataFrame.

    Parameters
    ----------
    spec : pandas.DataFrame
        A spec for an ActivitySim model.
    x_col: str
        The name of the columns in spec representing the data.
    p_col: dict
        The name of the columns in spec representing the parameters.
        The keys give the names of the columns to use, and the
        values will become the keys of the output dictionary.
    ignore_x : Collection, optional
        Labels in the spec file to ignore.  Typically this
        includes variables that are pre-processed by ActivitySim
        and therefore don't need to be made available in Larch.
    segment_id : str, optional
        The CHOOSER_SEGMENT_COLUMN_NAME identified for ActivitySim.
        This value is ignored if `p_col` is a string, and required
        if `p_col` is a dict.

    Returns
    -------
    dict
    """
    utils = {}
    for altname, altcode in p_col.items():
        utils[altcode] = linear_utility_from_spec(
            spec, x_col, altname, ignore_x=ignore_x
        )
    return utils


def explicit_value_parameters_from_spec(spec, p_col, model):
    """
    Define and lock parameters given as fixed values in the spec.

    Parameters
    ----------
    spec : pandas.DataFrame
        A spec for an ActivitySim model.
    p_col : str or dict
        The name of the columns in spec representing the parameters.
        Give as a string for a single column, or as a dict to
        have segments on multiple columns. If given as a dict,
        the keys give the names of the columns to use, and the
        values give the identifiers that will need to match the
        loaded `segment_id` value.  Only the keys are used in this
        function.
    model : larch.Model
        The model to insert fixed value parameters.

    Returns
    -------

    """
    if isinstance(p_col, dict):
        for p_col_ in p_col:
            explicit_value_parameters_from_spec(spec, p_col_, model)
    else:
        for i in spec.itertuples():
            try:
                j = float(getattr(i, p_col))
            except Exception:
                pass
            else:
                model.set_value(
                    getattr(i, p_col), value=j, holdfast=True,
                )


def explicit_value_parameters(model):
    """
    Define and lock parameters given as fixed values.

    Parameters
    ----------
    model : larch.Model
        The model to insert fixed value parameters.

    Returns
    -------

    """
    for i in model.pf.index:
        try:
            j = float(i)
        except Exception:
            pass
        else:
            model.set_value(
                i,
                value=j,
                initvalue=j,
                nullvalue=j,
                minimum=j,
                maximum=j,
                holdfast=True,
            )


def apply_coefficients(coefficients, model, minimum=None, maximum=None):
    """
    Read the coefficients CSV file to a DataFrame and set model parameters.

    Parameters
    ----------
    coefficients : pandas.DataFrame
        The coefficients table in the ActivitySim data bundle
        for this model.
    model : Model
        Apply coefficient values and constraints to this model.

    """
    if isinstance(model, dict):
        for m in model.values():
            apply_coefficients(coefficients, m)
    else:
        assert isinstance(coefficients, pd.DataFrame)
        assert all(coefficients.columns == ["value", "constrain"])
        assert coefficients.index.name == "coefficient_name"
        assert isinstance(model, AbstractChoiceModel)
        explicit_value_parameters(model)
        for i in coefficients.itertuples():
            if i.Index in model:
                model.set_value(
                    i.Index,
                    value=i.value,
                    holdfast=(i.constrain == "T"),
                    minimum=minimum,
                    maximum=maximum,
                )


def apply_coef_template(linear_utility, template_col, condition=None):
    """
    Apply a coefficient template over a linear utility function.

    Parameters
    ----------
    linear_utility : LinearFunction_C
    template_col : Mapping
    condition : any

    Returns
    -------
    LinearFunction_C
    """
    result = sum(
        P("*".join(template_col.get(ip, ip) for ip in i.param.split("*")))
        * i.data
        * i.scale
        for i in linear_utility
    )
    if condition is not None:
        result = result * condition
    return result


def construct_nesting_tree(alternatives, nesting_settings):
    """
    Construct a NestingTree from ActivitySim settings.

    Parameters
    ----------
    alternatives : Mapping or Sequence
        If given as a Mapping (dict), the keys are the alternative names
        as strings, and the values are alternative code numbers to use
        in larch.  If given as a Sequence, the values are the alternative
        names, and unique sequential codes will be created starting from 1.
    nesting_settings : Mapping
        The 'NESTS' section of the ActivitySim config file.

    Returns
    -------
    NestingTree
    """
    if not isinstance(alternatives, Mapping):
        alt_names = list(alternatives)
        alt_codes = np.arange(1, len(alt_names) + 1)
        alternatives = dict(zip(alt_names, alt_codes))

    tree = NestingTree()
    nest_names_to_codes = alternatives.copy()
    nest_names_to_codes["root"] = 0
    for alt_name, alt_code in alternatives.items():
        tree.add_node(alt_code, name=alt_name)

    def make_nest(cfg, parent_code=0):
        nonlocal nest_names_to_codes
        if cfg["name"] != "root":
            if cfg["name"] not in nest_names_to_codes:
                n = tree.new_node(
                    name=cfg["name"],
                    parameter=str(cfg["coefficient"]),
                    parent=parent_code,
                )
                nest_names_to_codes[cfg["name"]] = n
            else:
                tree.add_edge(parent_code, nest_names_to_codes[cfg["name"]])
        for a in cfg["alternatives"]:
            if isinstance(a, str):
                tree.add_edge(nest_names_to_codes[cfg["name"]], nest_names_to_codes[a])
            else:
                make_nest(a, parent_code=nest_names_to_codes[cfg["name"]])

    make_nest(nesting_settings)

    return tree



import numpy
from .exceptions import NoKnownShape, NoKnownType
from ..util.text_manip import truncate_path_for_display
from .general import selector_len_for
import uuid

class Pod():
	"""
	This is the abstract base class for Pods.
	"""

	def __init__(self, ident=None):
		if ident is None:
			self.ident = uuid.uuid1()
		else:
			self.ident = ident

	def load_data_item(self, name, result, selector=None):
		"""
		Load data into an existing array.

		The existing array must be the correct shape to receive the data. Errors can
		occur if the existing array is the incorrect shape.

		Parameters
		----------
		name : str
			The identifier for the data that will be loaded. This can be
			the natural name of a data item in this pod, or some expression
			that can be evaluated. The exact form of allowable expressions is
			dependent on the particular Pod implementation.
		result : ndarray
			This is the array (typically a slice of an array)
			into which the data will be loaded. It must be given and it
			must be the correct shape.
		selector : slice or array-like, optional
			This will slice the first dimension (cases) of the result.
			There may be a performace hit for using slices.

		Raises
		------
		NameError
			This pod cannot independently resolve the given name.
		"""
		raise NotImplementedError("overload this function")

	def _shape_for_get_data_item(self, selector=None):
		return [selector_len_for(selector, self.shape[0]), *self.shape[1:]]

	def get_data_item(self, name, selector=None, dtype=None):
		"""
		Get data in a new array.

		Parameters
		----------
		name : str
			The identifier for the data that will be loaded. This can be
			the natural name of a data item in this pod, or some expression
			that can be evaluated. The exact form of allowable expressions is
			dependent on the particular Pod implementation.
		selector : slice or array-like, optional
			This will slice the first dimension (cases) of the result.
			There may be a performace hit for using slices.
		dtype : dtype, optional
			The dtype for the array to return. If the dtype is not given and not
			inferable from the data itself, float64 will be used.

		Returns
		-------
		result : ndarray
			This is the array (typically a slice of an array)
			into which the data has beem loaded.

		Raises
		------
		NameError
			This pod cannot independently resolve the given name.
		"""
		try:
			natural_dtype = self.dtype_of(name)
		except:
			natural_dtype = numpy.float64
		result = numpy.zeros(self._shape_for_get_data_item(selector), dtype=dtype or natural_dtype)
		self.load_data_item(name, result, selector=selector)
		return result

	def get_data_dictionary(self, name):
		raise NotImplementedError("overload this function")

	@property
	def podtype(self):
		raise NotImplementedError("overload this function")

	@property
	def dims(self):
		"""The number of dimensions in a full data item"""
		raise NotImplementedError("overload this function")

	@property
	def shape(self):
		"""The shape of a full data item available in this pod"""
		raise NotImplementedError("overload this function")

	@shape.setter
	def shape(self, x):
		raise NotImplementedError("overload this function")

	@property
	def n_cases(self):
		"""The size of the first dimension in a full data item"""
		return self.shape[0]

	def names(self):
		"""Natural names of data items available in this pod"""
		raise NotImplementedError("overload this function")

	def nameset(self):
		"""Natural names of data items available in this pod, as a set"""
		n = self.names()
		if isinstance(n, set):
			return n
		return set(n)

	def __contains__(self, item):
		return (item in self.nameset())

	def dtype_of(self, name):
		"""dtype of raw data for a particular named data item"""
		raise NotImplementedError("overload this function")

	@property
	def filename(self):
		"""The filename of the underlying file (read-only)"""
		raise NotImplementedError("overload this function")

	@property
	def internalname(self):
		raise NotImplementedError("overload this function")


	def __repr__(self):
		from ..util.text_manip import max_len
		s = f"<larch.data_services.{self.__class__.__name__}>"
		try:
			filename = self.filename
		except NotImplementedError:
			pass
		else:
			s += f"\n |  file: {truncate_path_for_display(filename)}"
		try:
			internalnametype, internalnamevalue = self.internalname
		except NotImplementedError:
			pass
		else:
			s += f"\n |  {internalnametype}: {internalnamevalue}"
		try:
			shape = self.shape
		except (AttributeError, NoKnownShape, NotImplementedError):
			pass
		else:
			s += f"\n |  shape: {shape}"
		names = self.names()
		if len(names):
			s += "\n |  data:"
			just = max_len(names)
			for i in names:
				try:
					node_dtype = self.dtype_of(i)
				except NoKnownType:
					node_dtype = "<no dtype>"
				s += "\n |    {0:{2}s} ({1})".format(i, node_dtype, just)
		else:
			s += "\n |  data: <empty>"
		return s

	@property
	def durable_mask(self):
		"""A mask identifier for this pod, used to optimize reloading data"""
		try:
			return self._loading_mask
		except AttributeError:
			return numpy.uint32(0)

	@durable_mask.setter
	def durable_mask(self, value):
		self._loading_mask = numpy.uint32(value)

	@property
	def ident(self):
		"""An identifier for this pod, typically a `str`"""
		return self._ident

	@ident.setter
	def ident(self, x):
		self._ident = str(x)





import tempfile, webbrowser, types, os, shutil



TemporaryBucket = []

def TemporaryBucketCleanUp():
	global TemporaryBucket
	for i in TemporaryBucket:
		try:
			os.remove(os.path.realpath(i.name))
		except PermissionError:
			try:
				shutil.rmtree(i.name)
			except:
				import traceback
				#traceback.print_exc()
		except:
			import traceback
			#traceback.print_exc()
	del TemporaryBucket

import atexit
atexit.register(TemporaryBucketCleanUp)

def _open_in_chrome_or_something(url):
	# MacOS
	if os.path.exists('/Applications/Google Chrome.app'):
		chrome_path = r'open -a /Applications/Google\ Chrome.app %s'
	# Windows
	elif os.path.exists(r'C:\Program Files (x86)\Google\Chrome\Application\chrome.exe'):
		chrome_path = r'C:\Program Files (x86)\Google\Chrome\Application\chrome.exe %s'
	# Linux
	elif os.path.exists('/usr/bin/google-chrome'):
		chrome_path = '/usr/bin/google-chrome %s'
	else:
		chrome_path = None
	if chrome_path is None:
		webbrowser.open(url)
	else:
		webbrowser.get(chrome_path).open(url)

# Trap __exit__ to ensure the file does not get
# deleted when used in a with statement; to keep temporary files open
# until intentionally cleaned up
def _exit_without_closing(self, *arg, **kwarg):
	return self.file.__exit__(*arg, **kwarg)


def TemporaryFile(suffix='', mode='w+', use_chrome=True):
	t = tempfile.NamedTemporaryFile(suffix=suffix,mode=mode,delete=False)
	t.__exit__ = types.MethodType( _exit_without_closing, t )
	if use_chrome:
		t.view = types.MethodType( lambda self: _open_in_chrome_or_something('file://'+os.path.realpath(self.name)), t )
	else:
		t.view = types.MethodType( lambda self: webbrowser.open('file://'+os.path.realpath(self.name)), t )
	global TemporaryBucket
	TemporaryBucket.append(t)
	return t


def _try_write(self, content):
	try:
		self.write_(content)
	except:
		try:
			self.write_(str(content))
		except:
			self.write_(str(content).encode('utf-8'))


def TemporaryHtml(style=None, *, nohead=False, mode='wb+', content=None, **tagheads):
	t = TemporaryFile(suffix='.html', mode=mode)
	if 'b' in mode:
		t.write_ = t.write
		t.write = lambda x: _try_write(t,x)
	if not nohead and (style or len(tagheads)>0):
		t.write("<head>")
		if style:
			t.write("<style>{}</style>".format(style))
		for tag, content in tagheads.items():
			t.write("<{0}>{1}</{0}>".format(tag.lower(),content))
		t.write("</head>")
	if content is not None:
		t.write(content)
		t.view()
	return t


def TemporaryGzipInflation(gzfile):
	t = TemporaryFile(mode='wb')
	import gzip
	with gzip.open(gzfile, 'rb') as previewfile:
		t.write(previewfile.read())
	t.flush()
	return t.name

def TemporaryZipInflation(zfile):
	t = TemporaryDirectory(common=False)
	from zipfile import ZipFile
	ZipFile(zfile).extractall(path=t)
	return t

_tempdir = None

def TemporaryDirectory(common=True):
	global _tempdir
	if common and _tempdir is not None:
		return _tempdir
	t = tempfile.mkdtemp()
	global TemporaryBucket
	from . import Dict
	TemporaryBucket.append(Dict(name=t))
	if common:
		_tempdir = t
	return t

def TemporaryCopy(sourcefile, spool=True):
	tdir = TemporaryDirectory()
	basename = os.path.basename(sourcefile)
	if spool:
		n = 0
		vbasename = basename
		while os.path.exists(os.path.join(tdir,vbasename)):
			n += 1
			basename_ = os.path.splitext(basename)
			vbasename = "{0}.{2}{1}".format( basename_[0], basename_[1], n)
		shutil.copy2(sourcefile,os.path.join(tdir,vbasename))
		return os.path.join(tdir,vbasename)
	else:
		shutil.copy2(sourcefile,tdir)
		return os.path.join(tdir, os.path.basename(sourcefile))






def general_inverse(a, small_eig=0.0001):
	"""Find the matrix inverse if possible, otherwise find the pseudo-inverse."""
	import numpy
	if not numpy.isfinite(a).all():
		raise ValueError("nonfinite values in array")
	min_eig = None
	try:
		eig = numpy.linalg.eigvalsh(a)
		min_eig = numpy.min(numpy.abs(eig))
		if min_eig < small_eig:
			raise numpy.linalg.linalg.LinAlgError()
		x = numpy.linalg.inv(a)
	except numpy.linalg.linalg.LinAlgError:
		if min_eig is not None and min_eig < small_eig:
			import warnings
			warnings.warn(f"minimum eig {min_eig} in general_inverse")
		import scipy.linalg
		x = scipy.linalg.pinvh(a)
	return x






import ast
import numpy



def inXd(a,*b,**c):
	return numpy.in1d(a,*b,**c).reshape(a.shape)


class AstWrapper(ast.NodeTransformer):

	def visit_Compare(self, node):
		"Change the `in` operator to numpy.in1d"
		if len(node.ops)==1 and isinstance(node.ops[0], ast.In):
			#in1d = ast.Attribute(value=ast.Name(id='numpy', ctx=ast.Load()), attr='in1d', ctx=ast.Load())
			in1d = ast.Name(id='inXd', ctx=ast.Load())
			return ast.Call(in1d, [node.left, node.comparators[0]], [])
		elif len(node.ops)==1 and isinstance(node.ops[0], ast.NotIn):
			#in1d = ast.Attribute(value=ast.Name(id='numpy', ctx=ast.Load()), attr='in1d', ctx=ast.Load())
			in1d = ast.Name(id='inXd', ctx=ast.Load())
			return ast.Call(in1d, [node.left, node.comparators[0]], [ast.keyword('invert',ast.NameConstant(True))])
		else:
			return node


def asterize(cmd, mode="eval"):
	tree = ast.parse(cmd, mode=mode)
	tree = AstWrapper().visit(tree)
	# Add lineno & col_offset to the nodes we created
	ast.fix_missing_locations(tree)
	co = compile(tree, "<ast>", mode)
	return co





import os
from pathlib import Path
import tables as tb
import numpy
import pandas
import logging
from ....util import Dict
from ....util.aster import asterize
from ....util.text_manip import truncate_path_for_display
from ... import _reserved_names_
from ...pod import Pod
from ...general import _sqz_same, selector_len_for
from .... import warning

class IncompatibleShape(ValueError):
	pass

class NoKnownShape(ValueError):
	pass


class CArray(tb.CArray):

	@property
	def DICTIONARY(self):
		return Dict( self._v_attrs.DICTIONARY )

	@DICTIONARY.setter
	def DICTIONARY(self, x):
		self._v_attrs.DICTIONARY = dict(x)

	@property
	def DESCRIPTION(self):
		return self._v_attrs.TITLE

	@DESCRIPTION.setter
	def DESCRIPTION(self, x):
		self._v_attrs.TITLE = str(x)

	@property
	def TITLE(self):
		return self._v_attrs.TITLE

	@TITLE.setter
	def TITLE(self, x):
		self._v_attrs.TITLE = str(x)


	def __repr__(self):
		r = super().__repr__()
		try:
			d = self.DICTIONARY
		except:
			return r
		r += "\n  dictionary := {\n      "
		r += repr(d).replace("\n","\n      ")
		r += "\n  }"
		return r

	def uniques(self, slicer=None, counts=False):
		if isinstance(slicer, (bool, int)) and counts is False:
			counts = bool(slicer)
			slicer = None
		if slicer is None:
			slicer = slice(None)
		action = self[slicer]
		len_action = len(action)
		try:
			action = action[~numpy.isnan(action)]
		except TypeError:
			num_nan = 0
		else:
			num_nan = len_action - len(action)
		if counts:
			x = numpy.unique(action, return_counts=counts)
			try:
				d = self.DICTIONARY
			except AttributeError:
				y = pandas.Series(x[1], x[0])
			else:
				y = pandas.Series(x[1], [(d[j] if j in d else j) for j in x[0]])
			if num_nan:
				y[numpy.nan] = num_nan
			return y
		if num_nan:
			numpy.append(action, numpy.nan)
		return numpy.unique(action)


class H5Pod(Pod):
	def __init__(self, filename=None, mode='a', groupnode=None, *,
	             h5f=None, inmemory=False, temp=False,
	             complevel=1, complib='zlib',
				 do_nothing=False,
				 ident=None,
				 shape=None,
	             ):

		super().__init__(ident=ident)

		if do_nothing:
			return

		if isinstance(filename, H5Pod):
			# Copy / Re-Class contructor
			x = filename
			self._groupnode = x._groupnode
			self._h5f_own = False
			return

		if isinstance(filename, tb.group.Group) and groupnode is None:
			# Called with just a group node, use it
			groupnode = filename
			filename = None

		if isinstance(filename, (str,Path)):
			filename = os.fspath(filename)
			if groupnode is None:
				groupnode = "/"

		if filename is None and mode=='a' and groupnode is None:
			groupnode = '/' # default constructor for temp obj

		if isinstance(groupnode, tb.group.Group):
			# Use the existing info from this group node, ignore all other inputs
			self._groupnode = groupnode
			filename = groupnode._v_file.filename
			mode = groupnode._v_file.mode
			self._h5f_own = False

		elif isinstance(groupnode, str):

			# apply expanduser to filename to allow for home-folder based filenames
			if isinstance(filename,str):
				filename = os.path.expanduser(filename)

			if filename is None:
				temp = True
				from ....util.temporaryfile import TemporaryFile
				self._TemporaryFile = TemporaryFile(suffix='.h5d')
				filename = self._TemporaryFile.name

			if h5f is not None:
				self._h5f_own = False
				self._groupnode = self._h5f.get_node(groupnode)
			else:
				kwd = {}
				if inmemory or temp:
					kwd['driver']="H5FD_CORE"
				if temp:
					kwd['driver_core_backing_store']=0
				if complevel is not None:
					kwd['filters']=tb.Filters(complib=complib, complevel=complevel)
				self._h5f_obj = tb.open_file(filename, mode, **kwd)
				self._h5f_own = True
				try:
					self._groupnode = self._h5f_obj.get_node(groupnode)
				except tb.NoSuchNodeError:
					if isinstance(groupnode, str):
						self._groupnode = self._h5f_obj._get_or_create_path(groupnode, True)
		else:
			raise ValueError('must give groupnode as `str` or `tables.group.Group`')

		self._recoverable = (filename, self._groupnode._v_pathname)

		if shape is not None:
			self.shape = shape

	@property
	def _h5f(self):
		return self._groupnode._v_file

	def __repr__(self):
		from ....util.text_manip import max_len
		s = f"<larch.{self.__class__.__name__}>"
		try:
			s += f"\n |  file: {truncate_path_for_display(self.filename)}"
			if self._groupnode._v_pathname != "/":
				s += f"\n |  node: {self._groupnode._v_pathname}"
			try:
				shape = self.shape
			except NoKnownShape:
				shape = None
			else:
				s += f"\n |  shape: {shape}"
			try:
				metashape = self.metashape
			except (NoKnownShape, AttributeError):
				pass
			else:
				if metashape != shape:
					s += f"\n |  metashape: {metashape}"
			if len(self._groupnode._v_children):
				s += "\n |  data:"
				just = max_len(self._groupnode._v_children.keys())
				for i in sorted(self._groupnode._v_children.keys()):
					try:
						node_dtype = self._groupnode._v_children[i].dtype
					except tb.NoSuchNodeError:
						node_dtype = "<no dtype>"
					s += "\n |    {0:{2}s} ({1})".format(i, node_dtype, just)
			else:
				s += "\n |  data: <empty>"
		except (tb.ClosedNodeError, tb.ClosedFileError):
			s += f"\n |  <file is closed>"
			s += f"\n |  file: {truncate_path_for_display(self._recoverable[0])}"
			s += f"\n |  node: {self._recoverable[1]}"
		return s

	def __xml__(self, no_data=False, descriptions=True, dictionaries=False):
		from xmle import Elem
		x = Elem('div')
		t = x.elem('table', style="margin-top:1px;")
		t.elem('caption', text=f"<larch.{self.__class__.__name__}>", style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;")
		#
		try:
			ident = self.ident
		except AttributeError:
			pass
		else:
			tr = t.elem('tr')
			tr.elem('th', text='ident')
			tr.elem('td', text=ident)
		#
		try:
			filename = self.filename
		except AttributeError:
			pass
		else:
			tr = t.elem('tr')
			tr.elem('th', text='file')
			tr.elem('td', text=truncate_path_for_display(filename))
		#
		try:
			filemode = self.filemode
		except AttributeError:
			pass
		else:
			tr = t.elem('tr')
			tr.elem('th', text='mode')
			tr.elem('td', text=truncate_path_for_display(self.filemode))
		#
		try:
			v_pathname = self._groupnode._v_pathname
		except AttributeError:
			pass
		else:
			if self._groupnode._v_pathname != "/":
				tr = t.elem('tr')
				tr.elem('th', text='node')
				tr.elem('td', text=self._groupnode._v_pathname)
		#
		try:
			str_shape = str(self.shape)
		except NoKnownShape:
			pass
		else:
			tr = t.elem('tr')
			tr.elem('th', text='shape')
			tr.elem('td', text=str_shape)
		#
		try:
			str_shape = str(self.metashape)
		except (NoKnownShape, AttributeError):
			pass
		else:
			tr = t.elem('tr')
			tr.elem('th', text='metashape')
			tr.elem('td', text=str_shape)
		#
		try:
			str_durable_mask = f"0x{self.durable_mask:X}"
		except (AttributeError):
			pass
		else:
			if str_durable_mask!='0x0':
				tr = t.elem('tr')
				tr.elem('th', text='durable_mask')
				tr.elem('td', text=str_durable_mask)
		#
		if not no_data:
			if len(self._groupnode._v_children):
				tr = t.elem('tr')
				tr.elem('th', text='data', style='vertical-align:top;')
				td = tr.elem('td')
				t1 = td.elem('table', cls='dictionary')
				t1head = t1.elem('thead')
				t1headr = t1head.elem('tr')
				t1headr.elem('th', text='name')
				t1headr.elem('th', text='dtype')
				if descriptions:
					t1headr.elem('th', text='description')
				any_sources = 0
				for i in sorted(self._groupnode._v_children.keys()):
					try:
						node_dtype = self._groupnode._v_children[i].dtype
					except (tb.NoSuchNodeError, AttributeError):
						node_dtype = "<no dtype>"
					if i not in _reserved_names_:
						tr1 = t1.elem('tr')
						tr1.elem('td', text=i)
						tr1.elem('td', text=node_dtype)
						if descriptions:
							try:
								title = self._groupnode._v_children[i]._v_attrs['TITLE']
							except:
								title = ""
							else:
								tr1.elem('td', text=title)
						try:
							orig_source = self._groupnode._v_children[i]._v_attrs['ORIGINAL_SOURCE']
						except:
							pass
						else:
							tr1.elem('td', text=orig_source)
							any_sources += 1
				if any_sources:
					t1headr.elem('th', text='source')
			else:
				tr = t.elem('tr')
				tr.elem('th', text='data', style='vertical-align:top;')
				tr.elem('td', text='<empty>')
		return x

	def _repr_html_(self):
		return self.__xml__().tostring()

	def change_mode(self, mode, **kwarg):
		"""Change the file mode of the underlying HDF5 file.

		Can be used to change from read-only to read-write.
		"""
		if mode == self.filemode:
			return
		if mode == 'w':
			raise TypeError("cannot change_mode to w, close the file and delete it")
		filename = self.filename
		groupnode_path = self._groupnode._v_pathname
		self.close()
		self.__init__(filename, mode, groupnode=groupnode_path, **kwarg)
		return self

	def reopen(self, mode='r', **kwarg):
		"""Reopen the underlying HDF5 file.

		Can be used to change from read-only to read-write or to reopen a file that was closed.
		"""
		if mode == self.filemode:
			return
		if mode == 'w':
			raise TypeError("cannot change_mode to w, close the file and delete it")
		filename = self.filename
		groupnode_path = self.groupnode_path
		try:
			self.close()
		except tb.ClosedNodeError:
			pass
		self.__init__(filename, mode, groupnode=groupnode_path, **kwarg)
		return self

	def names(self):
		return [i for i in self._groupnode._v_children.keys() if i not in _reserved_names_]

	def rename_vars(self, *rename_vars):
		"""
		Rename variables according to patterns.

		Parameters
		----------
		rename_vars : 2-tuples
			A sequence of 2-tuples, giving (pattern, replacement) that will be fed to re.sub.
			For example, give ('^','prefix_') to add prefix to all variable names, or
			('^from_this_name$','to_this_name') to change an exact name from one thing to another.

		"""
		import re
		for pattern, replacement in rename_vars:
			q = [(re.sub(pattern, replacement, k),k) for k in self.names()]
			for _to,_from in q:
				if _to != _from:
					self._groupnode._v_children[_from]._f_rename(_to)

	def reshape(self, *shape):
		if len(shape)==0:
			raise ValueError('no shape given')
		if len(shape)==1 and isinstance(shape[0], tuple):
			shape = shape[0]
		if isinstance(shape, int):
			shape = (shape,)
		if not isinstance(shape, tuple):
			raise TypeError('reshape must be int or tuple')
		if len(shape)==1 and shape[0] == -1:
			shape = (int(numpy.product(self.shape)), )
		elif len(shape)==2:
			if shape[0] == -1 and shape[1] > 0:
				shape = (int(numpy.product(self.shape) / shape[1]), shape[1])
			if shape[1] == -1 and shape[0] > 0:
				shape = (shape[0], int(numpy.product(self.shape) / shape[0]))
		if numpy.product(shape) != numpy.product(self.shape):
			raise ValueError(f'incompatible reshape {shape} for current shape {self.shape}')
		#print("reshape to", shape)
		for k in self._groupnode._v_children.keys():
			#print("reshape",k,shape)
			self._groupnode._v_children[k].shape = shape
		self.shape = shape

	def __dir__(self):
		x = super().__dir__()
		x.extend(self.names())
		return x

	@property
	def filename(self):
		try:
			return self._groupnode._v_file.filename
		except (tb.ClosedNodeError, tb.ClosedFileError) as err:
			try:
				return self._last_closed_filename
			except AttributeError:
				raise err

	@property
	def filemode(self):
		try:
			return self._groupnode._v_file.mode
		except (tb.ClosedNodeError, tb.ClosedFileError) as err:
			return None

	@property
	def groupnode_path(self):
		try:
			return self._groupnode._v_pathname
		except (tb.ClosedNodeError, tb.ClosedFileError) as err:
			try:
				return self._last_closed_groupnode_path
			except AttributeError:
				raise err

	@property
	def n_cases(self):
		return self.shape[0]

	@property
	def shape(self):
		"""The shape of the pod.

		"""
		if 'SHAPE' in self._groupnode._v_attrs:
			return tuple(self._groupnode._v_attrs['SHAPE'][:])
		if len(self.names()):
			for v in self._groupnode._v_children.values():
				try:
					found_shape = v.shape
				except:
					pass
				else:
					try:
						self.shape = found_shape
					except:
						pass
					return tuple(found_shape)
		raise NoKnownShape()

	@shape.setter
	def shape(self, x):
		# if self._groupnode._v_nchildren > 0:
		# 	raise ValueError('this pod has shape {!s} but you want to set {!s}'.format(self.shape, x))
		# if self._groupnode._v_nchildren == 0:
		self._groupnode._v_attrs.SHAPE = numpy.asarray(x, dtype=int)

	@property
	def metashape(self):
		"""The actual shape of the data underlying the pod, often same as shape."""
		return self.shape

	def add_expression(self, name, expression, *, overwrite=False, title=None, dictionary=None, dtype=None):
		arr = self[expression]
		if dtype is not None:
			arr = arr.astype(dtype)
			try:
				dtype_str = "("+dtype.__name__+")"
			except:
				dtype_str = ""
			original_source = f'={dtype_str} {expression}'
		else:
			original_source = f'= {expression}'
		if overwrite=='ignore':
			overwrite = False
			if_exists = 'ignore'
		else:
			if_exists = 'raise'
		try:
			self.add_array(name, arr, overwrite=overwrite, title=title, dictionary=dictionary,
						   original_source=original_source, rel_original_source=False)
		except tb.exceptions.NodeError:
			if if_exists=='ignore':
				pass
			else:
				raise

	def add_array(self, name, arr, *, overwrite=False, original_source=None, rel_original_source=True,
	              title=None, dictionary=None, fix_name_problems=True):
		"""Create a new variable in the H5Pod.

		Creating a new variable in the data might be convenient in some instances.
		If you create an array externally, you can add it to the file easily with
		this command.

		Parameters
		----------
		name : str
			The name of the new variable.
		arr : ndarray
			An array to add as the new variable.  Must have the correct shape.
		overwrite : bool
			Should the variable be overwritten if it already exists, default to False.
		original_source : str
			Optionally, give the file name or other description of the source of the data in this array.
		rel_original_source : bool
			If true, change the absolute path of the original_source to a relative path viz this file.
		title : str, optional
			A descriptive title for the variable, typically a short phrase but an
			arbitrary length description is allowed.
		dictionary : dict, optional
			A data dictionary explaining some or all of the values in this field.
			Even for otherwise self-explanatory numerical values, the dictionary
			may give useful information about particular out of range values.

		Raises
		------
		tables.exceptions.NodeError
			If a variable of the same name already exists.
		"""
		if name in _reserved_names_:
			raise ValueError(f'{name} is a reserved name')
		if '/' in name and fix_name_problems:
			import warnings
			warnings.warn(f'the ``/`` character is not allowed in variable names ({name})\n'
						  f'changing it to ``|``')
			name = name.replace('/','|')
		try:
			existing_shape = tuple(self.metashape)
		except NoKnownShape:
			pass
		else:
			if existing_shape != arr.shape:
				# maybe just has extra size-1 dims, check for that...
				arr = arr.squeeze()
				if self.podtype == 'idcs':
					if existing_shape[:-1] != arr.shape:
						raise IncompatibleShape(
							"new array must have shape {!s} but the array given has shape {!s}".format(self.shape, arr.shape))
				else:
					if existing_shape != arr.shape:
						raise IncompatibleShape(
							"new array must have shape {!s} but the array given has shape {!s}".format(self.shape, arr.shape))
		if overwrite:
			self.delete_array(name)
		try:
			h5var = self._h5f.create_carray(self._groupnode, name, obj=arr)
		except ValueError as valerr:
			if "unknown type" in str(valerr) or "unknown kind" in str(valerr):  # changed for pytables 3.3
				try:
					tb_atom = tb.Atom.from_dtype(arr.dtype)
				except ValueError:
					log = logging.getLogger('H5')
					try:
						maxlen = int(len(max(arr.astype(str), key=len)))
					except ValueError:
						import datetime
						if 0:  # isinstance(arr[0], datetime.time):
							log.warning(f"  column {name} is datetime.time, converting to Time32")
							tb_atom = tb.atom.Time32Atom()

							# convert_datetime_time_to_epoch_seconds = lambda tm: tm.hour*3600+ tm.minute*60 + tm.second
							def convert_datetime_time_to_epoch_seconds(tm):
								try:
									return tm.hour * 3600 + tm.minute * 60 + tm.second
								except:
									if numpy.isnan(tm):
										return 0
									else:
										raise

							arr = arr.apply(convert_datetime_time_to_epoch_seconds).astype(numpy.int32).values
						else:
							# import __main__
							# __main__.err_df = df
							raise
					else:
						maxlen = max(maxlen, 8)
						if arr.dtype != object:
							log.warning(f"cannot create column {name} as dtype {arr.dtype}, converting to S{maxlen:d}")
						arr = arr.astype('S{}'.format(maxlen))
						tb_atom = tb.Atom.from_dtype(arr.dtype)
				h5var = self._h5f.create_carray(self._groupnode, name, tb_atom, shape=arr.shape)
				h5var[:] = arr
			else:
				raise
		if rel_original_source and original_source and original_source[0] != '=':
			basedir = os.path.dirname(self.filename)
			original_source = os.path.relpath(original_source, start=basedir)
		if original_source is not None:
			h5var._v_attrs.ORIGINAL_SOURCE = original_source
		if title is not None:
			h5var._v_attrs.TITLE = title
		if dictionary is not None:
			h5var._v_attrs.DICTIONARY = dictionary

	def add_blank(self, name, shape=None, dtype=numpy.float64, **kwargs):
		"""Create a new variable in the H5Pod.

		Creating a new variable in the data might be convenient in some instances.
		If you create an array externally, you can add it to the file easily with
		this command.

		Parameters
		----------
		name : str
			The name of the new variable.
		dtype : dtype
			The dtype of the empty array to add as the new variable.
		shape : tuple
			The shape of the empty array to add. Must be compatible with existing
			shape, if any.

		Other keyword parameters are passed through to `add_array`.

		Raises
		------
		tables.exceptions.NodeError
			If a variable of the same name already exists.
		NoKnownShape
			If shape is not given and not already known from the file.
		"""
		if name in _reserved_names_:
			raise ValueError(f'{name} is a reserved name')
		try:
			existing_shape = tuple(self.metashape)
		except NoKnownShape:
			if shape is None:
				raise
		else:
			if shape is None:
				shape = existing_shape
			if existing_shape != tuple(shape):
				raise IncompatibleShape(
					"new array must have shape {!s} but the array given has shape {!s}".format(self.shape, shape))
		return self.add_array(name, numpy.zeros(shape, dtype=dtype), **kwargs)


	def delete_array(self, name, recursive=True):
		"""Delete an existing variable.

		Parameters
		----------
		name : str
			The name of the data node to remove.
		recursive : bool
			If the data node is a group, recursively remove all sub-nodes.

		"""
		if name in _reserved_names_:
			raise ValueError(f'{name} is a reserved name')
		try:
			self._h5f.remove_node(self._groupnode, name, recursive)
		except tb.exceptions.NoSuchNodeError:
			pass


	def flush(self, *arg, **kwargs):
		return self._h5f.flush(*arg, **kwargs)

	def close(self, *arg, **kwargs):
		try:
			self._last_closed_filename = self.filename
			self._last_closed_groupnode_path = self.groupnode_path
		except:
			pass
		return self._h5f.close(*arg, **kwargs)

	@property
	def podtype(self):
		return ''

	def uri(self, mode=None):
		from urllib.parse import urlunparse
		q_dict = {}
		if self.podtype:
			q_dict['type'] = self.podtype
		if mode:
			q_dict['mode'] = mode
		q = "&".join(f'{k}={v}' for k,v in q_dict.items())
		return urlunparse(['file', '', self.filename, '', q, self._groupnode._v_pathname])

	def _remake_command(self, cmd, selector=None, receiver=None):
		from tokenize import tokenize, untokenize, NAME, OP, STRING
		DOT = (OP, '.')
		COLON = (OP, ':')
		COMMA = (OP, ',')
		OBRAC = (OP, '[')
		CBRAC = (OP, ']')
		OPAR = (OP, '(')
		CPAR = (OP, ')')
		from io import BytesIO
		recommand = []

		if receiver:
			recommand += [(NAME, receiver), OBRAC, COLON, CBRAC, (OP, '='), ]

		try:
			cmd_encode = cmd.encode('utf-8')
		except AttributeError:
			cmd_encode = str(cmd).encode('utf-8')
		dims = len(self.shape)
		g = tokenize(BytesIO(cmd_encode).readline)
		if selector is None:
			screen_tokens = [COLON,]
		else:
			# try:
			# 	slicer_encode = selector.encode('utf-8')
			# except AttributeError:
			# 	slicer_encode = str(selector).encode('utf-8')
			# screen_tokens = [(toknum, tokval) for toknum, tokval, _, _, _ in tokenize(BytesIO(slicer_encode).readline)]
			screen_tokens = [(NAME, 'selector'), ]
		for toknum, tokval, _, _, _ in g:
			if toknum == NAME and tokval in self._groupnode:
				# replace NAME tokens
				partial = [(NAME, 'self'), DOT, (NAME, '_groupnode'), DOT, (NAME, tokval), OBRAC, ]
				partial += screen_tokens
				if len(self._groupnode._v_children[tokval].shape)>1:
					partial += [COMMA, COLON, ]
				if len(self._groupnode._v_children[tokval].shape)>2:
					partial += [COMMA, COLON, ]
				if len(self._groupnode._v_children[tokval].shape)>3:
					partial += [COMMA, COLON, ]
				partial += [CBRAC,]


				recommand.extend(partial)
			else:
				recommand.append((toknum, tokval))
		# print("<recommand>")
		# print(recommand)
		# print("</recommand>")
		ret = untokenize(recommand).decode('utf-8')
		return asterize(ret, mode="exec" if receiver is not None else "eval"), ret


	def _evaluate_single_item(self, cmd, selector=None, receiver=None):
		j, j_plain = self._remake_command(cmd, selector=selector, receiver='receiver' if receiver is not None else None)
		# important globals
		from ....util.aster import inXd
		from numpy import log, exp, log1p, absolute, fabs, sqrt, isnan, isfinite, logaddexp, fmin, fmax, nan_to_num, sin, cos, pi
		from ....util.common_functions import piece, normalize, boolean
		try:
			if receiver is not None:
				exec(j)
			else:
				return eval(j)
		except Exception as exc:
			args = exc.args
			if not args:
				arg0 = ''
			else:
				arg0 = args[0]
			arg0 = arg0 + '\nwithin parsed command: "{!s}"'.format(cmd)
			arg0 = arg0 + '\nwithin re-parsed command: "{!s}"'.format(j_plain)
			if selector is not None:
				arg0 = arg0 + '\nwith selector: "{!s}"'.format(selector)
			if "max" in cmd:
				arg0 = arg0 + '\n(note to get the maximum of arrays use "fmax" not "max")'.format(cmd)
			if "min" in cmd:
				arg0 = arg0 + '\n(note to get the minimum of arrays use "fmin" not "min")'.format(cmd)
			if isinstance(exc, NameError):
				badname = str(exc).split("'")[1]
				goodnames = dir()
				from ....util.text_manip import case_insensitive_close_matches
				did_you_mean_list = case_insensitive_close_matches(badname, goodnames, n=3, cutoff=0.1, excpt=None)
				if len(did_you_mean_list) > 0:
					arg0 = arg0 + '\n' + "did you mean {}?".format(
						" or ".join("'{}'".format(s) for s in did_you_mean_list))
			exc.args = (arg0,) + args[1:]
			raise

	def __contains__(self, item):
		if item in self._groupnode:
			return True
		return False

	def dtype_of(self, name):
		"""dtype of raw data for a particular named data item."""
		if name in self._groupnode._v_children:
			return self._groupnode._v_children[name].dtype
		raise KeyError(f"{name} not found")

	def load_data_item(self, name, result, selector=None):
		"""Load a slice of the pod arrays into an array in memory"""
		# convert a single name string to a one item list
		_sqz_same(result.shape, [selector_len_for(selector, self.shape[0]), *self.shape[1:]])
		try:
			result[:] = self._evaluate_single_item(name, selector)
		except IndexError:
			# https://github.com/PyTables/PyTables/issues/310
			_temp = self._evaluate_single_item(name, None)
			try:
				result[:] = _temp[selector]
			except Exception as err:
				raise ValueError(f'_temp.shape={_temp.shape}  selector.shape={selector.shape}') from err
		return result

	def load_meta_data_item(self, name, result, selector=None):
		"""Load a slice of the pod arrays into an array in memory"""

		if selector is not None:
			import warnings
			warnings.warn('selector not compatible for load_meta_data_item')

		# convert a single name string to a one item list
		_sqz_same(result.shape, self.metashape)
		try:
			result[:] = self._evaluate_single_item(name, selector)
		except IndexError:
			# https://github.com/PyTables/PyTables/issues/310
			result[:] = self._evaluate_single_item(name, None)[selector]
		return result

	def get_data_dictionary(self, name):
		"""dictionary of raw data for a particular named data item."""
		if name in self._groupnode._v_children:
			return self._groupnode._v_children[name].DICTIONARY
		raise KeyError(f"{name} not found")

	def __getitem__(self, item):
		if isinstance(item, tuple) and len(item)>=2 and isinstance(item[-1], slice):
			names, slice_ = item[:-1], item[-1]
		else:
			names = item
			slice_ = None

		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names,]

		dtype = numpy.float64

		result = numpy.zeros( [selector_len_for(slice_, self.shape[0]), *self.shape[1:], len(names)], dtype=dtype)

		for i, cmd in enumerate(names):
			result[...,i] = self._evaluate_single_item(cmd, slice_)

		return result

	def _load_into(self, names, slc, result):
		"""Load a slice of the pod arrays into an array in memory"""
		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]
		_sqz_same(result.shape,[selector_len_for(slc, self.shape[0]), *self.shape[1:], len(names)])
		for i, cmd in enumerate(names):
			result[..., i] = self._evaluate_single_item(cmd, slc)
		return result

	def load_into(self, names, selector, result):
		"""Load a slice of the pod arrays into an array in memory"""
		# convert a single name string to a one item list
		if isinstance(names, str):
			names = [names, ]
		_sqz_same(result.shape, [selector_len_for(selector, self.shape[0]), *self.shape[1:], len(names)])
		for i, cmd in enumerate(names):
			result[..., i] = self._evaluate_single_item(cmd, selector)
		return result

	def __getattr__(self, item):
		if item in self._groupnode._v_children:
			ret = self._groupnode._v_children[item]
			if isinstance(ret, tb.CArray):
				ret.__class__ = CArray
			return ret
		raise AttributeError(item)

	class _dataframe_factory():
		def __init__(self, obj):
			self.obj = obj
		def __getattr__(self, item):
			return getattr(self.obj,item)
		def __getitem__(self, item):
			if len(self.obj.shape) > 1:
				try:
					metashape = self.obj.metashape
				except AttributeError:
					raise TypeError('dataframe access currently only compatible with 1d, use regular arrays for higher dimensions')
				else:
					if len(metashape) > 1:
						raise TypeError('dataframe access currently only compatible with 1d, use regular arrays for higher dimensions')

			if isinstance(item, tuple) and len(item)>=2 and isinstance(item[-1], slice):
				names, slice_ = item[:-1], item[-1]
			else:
				names = item
				slice_ = None

			# convert a single name string to a one item list
			if isinstance(names, str):
				names = [names,]

			result = pandas.DataFrame()

			for i, cmd in enumerate(names):
				j = self.obj._evaluate_single_item(cmd, selector=slice_)
				try:
					#result.loc[:,cmd] = j
					result = result.assign(**{cmd:j})
				except:
					print()
					print(f"An error in tacking {cmd} to result")
					print(f"j.dtype is {j.dtype}")
					print(f"j.shape is {j.shape}")
					print(f"result.shape is {result.shape}")
					print()
					raise

			return result

	@property
	def dataframe(self):
		return self._dataframe_factory(self)

	def astype(self, t:str):
		from . import _pod_types
		cls = _pod_types[t.lower()]
		return cls(self)

	def statistics_for(self, var, histogram=True, selector=None, **kwargs):
		a = self.get_data_item(var)
		if isinstance(selector, str):
			selector = self.get_data_item(selector, None, dtype=bool)
		if selector is not None:
			a = a[selector]
		from ....util.statistics import statistics_for_array
		try:
			dictionary = self._groupnode._v_children[var]._v_attrs.DICTIONARY
		except:
			dictionary = None
		try:
			descrip = self._groupnode._v_children[var]._v_attrs.TITLE
		except:
			descrip = None
		result = statistics_for_array(a, histogram=histogram, varname=var, dictionary=dictionary, **kwargs)
		if descrip is not None and descrip!="":
			result.description = descrip
		if dictionary is not None:
			result.dictionary = Dict(dictionary)
		return result

	def statistics(self, vars=None, histogram=False, selector=None):
		if vars is None:
			vars = self.names()
		from ....util import Dict
		from ....util.arraytools import scalarize
		import numpy.ma as ma
		stats = pandas.DataFrame(
			columns=[
				'n',
				'minimum',
				'maximum',
				'median',
				'mean',
				'stdev',
				'nonzero_minimum',
				'nonzero_maximum',
				'nonzero_mean',
				'nonzero_stdev',
				'zeros',
				'positives',
				'negatives',
			] + (['histogram'] if histogram else []),
			index = vars,
		)
		for var in vars:
			if selector is not None:
				if isinstance(selector, slice):
					a = self[var, selector]
				else:
					a = self[var][selector]
			else:
				a = self[var]
			stats.loc[var,'n'] = scalarize(a.shape[0])
			stats.loc[var,'minimum'] = scalarize(numpy.nanmin(a, axis=0))
			stats.loc[var,'maximum'] = scalarize(numpy.nanmax(a, axis=0))
			stats.loc[var,'median'] = scalarize(numpy.nanmedian(a, axis=0))
			if histogram:
				from ....util.histograms import sizable_histogram_figure, seems_like_discrete_data
				try:
					dictionary = self.get_data_dictionary(var)
				except:
					dictionary = None
				stats.loc[var,'histogram'] = sizable_histogram_figure(
					a,
					title=None, xlabel=var, ylabel='Frequency',
					discrete=seems_like_discrete_data(a, dictionary)
					)
			ax = ma.masked_array(a, mask=~numpy.isfinite(a))
			stats.loc[var,'mean'] = scalarize(numpy.mean(ax, axis=0))
			stats.loc[var,'stdev'] = scalarize(numpy.std(ax, axis=0))
			stats.loc[var, 'zeros'] = scalarize(numpy.sum(numpy.logical_not(ax), axis=0))
			stats.loc[var, 'positives'] = scalarize(numpy.sum(ax>0, axis=0))
			stats.loc[var, 'negatives'] = scalarize(numpy.sum(ax<0, axis=0))
			ax.mask |= (ax==0)
			stats.loc[var,'nonzero_minimum'] = scalarize(numpy.min(ax, axis=0))
			stats.loc[var,'nonzero_maximum'] = scalarize(numpy.max(ax, axis=0))
			stats.loc[var,'nonzero_mean'] = scalarize(numpy.mean(ax, axis=0))
			stats.loc[var,'nonzero_stdev'] = scalarize(numpy.std(ax, axis=0))
		if histogram:
			from ....util.dataframe import DataFrameViewer
			return DataFrameViewer(stats)
		return stats

	def get_row(self, rownum, lookup=True):
		result = Dict()
		for i in self.names():
			result[i] = self._groupnode._v_children[i][rownum]
			if lookup:
				try:
					d = self._groupnode._v_children[i]._v_attrs.DICTIONARY
				except (KeyError, AttributeError):
					pass
				else:
					if result[i] in d:
						result[i] = f"{result[i]} ({d[result[i]]})"
		return result

	@property
	def vault(self):
		try:
			return self.__vault
		except:
			from ..h5util import get_or_create_group
			from ..h5vault import H5Vault
			v = get_or_create_group( self._h5f, self._groupnode, name='_VAULT_', title='', filters=None, createparents=False, skip_on_readonly=False )
			self.__vault = H5Vault(v)
			return self.__vault

